#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Andrew Klychkov (@Andersson007) <andrew.a.klychkov@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
---
module: clickhouse_grants

short_description: Manage grants for ClickHouse users and roles

description:
  - Grants, updates, or revokes privileges for ClickHouse users and roles.
  - This module uses the L(clickhouse-driver,https://clickhouse-driver.readthedocs.io/en/latest) client interface.

attributes:
  check_mode:
    description: Supports check_mode.
    support: full
  idempotent:
    description: At second run will not change anything.
    support: full

author:
  - Andrew Klychkov (@Andersson007)
  - Fabian Kohn (@fako1024)

extends_documentation_fragment:
  - community.clickhouse.client_inst_opts

version_added: '0.9.0'

options:
  state:
    description:
      - If C(present), the module will grant or update privileges.
      - If C(absent), the module will revoke all privileges from the O(grantee).
    type: str
    choices: ['present', 'absent']
    default: 'present'
  grantee:
    description:
      - A user or a role to grant, update, or revoke privileges for.
    type: str
    required: true
  exclusive:
    description:
      - If set to C(false) (the default), the module will append
        the privileges specified in O(privileges) to the privileges the O(grantee)
        already has.
      - If set to C(true), the module will revoke all
        current privileges from the O(grantee) before granting the new ones.
    type: bool
    default: false
  privileges:
    description:
      - Privileges to grant. This option is required when C(state) is C(present).
      - It's a list of dictionaries, where each dictionary specifies a set of privileges on a database object.
    type: list
    elements: dict
    suboptions:
      object:
        description:
          - The database object to grant privileges on.
          - Use C(*.*) for global privileges, C(database.*) for all tables in a database,
            and C(database.table) for a specific table.
        type: str
        required: true
      privs:
        description:
          - A dictionary of privileges.
          - Keys are privilege names, like C(CREATE USER) or V(SELECT(column1, column2\)).
          - Values are booleans indicating whether to grant the privilege
            with the C(WITH GRANT OPTION).
          - Alternatively, you can use the C(grant_option) parameter to apply the same setting to all privileges in this set.
        type: dict
        required: true
      grant_option:
        description:
          - A boolean that applies to all privileges in this set.
          - If specified, it overrides any individual grant option settings within C(privs).
        type: bool
  cluster:
    description:
      - Run the grant/revoke commands on all cluster hosts.
      - If the cluster is not configured, the command will fail with an error.
    type: str
    version_added: '0.11.0'
'''

EXAMPLES = r'''
- name: Grant global privileges to a user
  community.clickhouse.clickhouse_grants:
    grantee: alice
    privileges:
      - object: '*.*'
        privs:
          "ALTER USER": true       # With grant option
          "CREATE DATABASE": false # Without grant option
          "CREATE USER": false     # Without grant option

- name: Grant privileges on a specific database
  community.clickhouse.clickhouse_grants:
    grantee: bob
    privileges:
      - object: 'infra.*'
        privs:
          "SELECT": true  # With grant option
          "INSERT": false # Without grant option

- name: Grant privileges on a cluster
  community.clickhouse.clickhouse_grants:
    grantee: bob
    cluster: test_cluster
    privileges:
      - object: 'infra.*'
        privs:
          "SELECT": true  # With grant option
          "INSERT": false # Without grant option

- name: Grant SELECT on specific columns of a table
  community.clickhouse.clickhouse_grants:
    grantee: carol
    privileges:
      - object: 'sales.customers'
        privs:
          "SELECT(name, email)": false # Without grant option

- name: Replace all existing privileges for a user
  community.clickhouse.clickhouse_grants:
    grantee: david
    exclusive: true
    privileges:
      - object: 'bar.*'
        privs:
          "SELECT": false  # Without grant option

- name: Revoke all privileges from a user
  community.clickhouse.clickhouse_grants:
    grantee: eve
    state: absent
'''

RETURN = r'''
executed_statements:
  description:
  - Data-modifying executed statements.
  returned: on success
  type: list
  sample: ['GRANT SELECT ON foo.* TO alice', 'REVOKE INSERT ON foo.* FROM alice']
diff:
  description:
  - Differences between the previous and current state.
  - Only returned when diff mode is enabled (with C(--diff) or in C(check_mode)).
  returned: when diff mode is enabled or check_mode is true
  type: dict
  contains:
    before:
      description: Grants before the change.
      returned: always
      type: dict
      sample:
        "*.*":
          "CREATE USER": false
        "foo.*":
          "INSERT": false
          "SELECT": true
    after:
      description: Grants after the change.
      returned: always
      type: dict
      sample:
        "*.*":
          "CREATE USER": false
        "foo.*":
          "DELETE": true
          "INSERT": false
          "SELECT": true
'''

import re
from collections import defaultdict

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    check_clickhouse_driver,
    client_common_argument_spec,
    connect_to_db_via_client,
    execute_query,
    get_main_conn_kwargs,
)

executed_statements = []

# Compile regex pattern once for performance
GRANT_REGEX = re.compile(r'GRANT (.+?) ON (.+?) TO .+?( WITH GRANT OPTION)?(?: ON CLUSTER .+)?$')


class ClickHouseGrants():
    def __init__(self, module, client, grantee, cluster=None):
        self.changed = False
        self.module = module
        self.client = client
        self.grantee = grantee
        self.cluster = cluster
        self.__check_grantee_exists()

    def __check_grantee_exists(self):
        # Check if grantee exists as either a user or a role
        query = ("SELECT 1 FROM system.users WHERE name = '%s' "
                 "UNION ALL "
                 "SELECT 1 FROM system.roles WHERE name = '%s' "
                 "LIMIT 1" % (self.grantee, self.grantee))

        result = execute_query(self.module, self.client, query)

        if not result:
            self.module.fail_json(msg="Grantee %s does not exist" % self.grantee)

    def get(self):
        query = "SHOW GRANTS FOR '%s'" % self.grantee
        result = execute_query(self.module, self.client, query)

        grants = {}
        for row in result:
            grant_statement = row[0]
            match = GRANT_REGEX.match(grant_statement)
            if not match:
                continue

            privs_str, obj, grant_option_str = match.groups()
            grant_option = (grant_option_str is not None)

            # Normalize object name: ClickHouse 25.x returns '*' instead of '*.*' for global grants
            if obj == '*':
                obj = '*.*'

            if obj not in grants:
                grants[obj] = {}

            privs = [p.strip().upper() for p in privs_str.split(',')]
            for priv in privs:
                grants[obj][priv] = grant_option

        return grants

    def _get_desired_grants(self):
        privileges = self.module.params['privileges']
        if not privileges:
            return {}

        desired_grants = {}
        for p in privileges:
            obj = p['object']
            if obj not in desired_grants:
                desired_grants[obj] = {}

            grant_option_override = p.get('grant_option')
            for priv, grant_option in p['privs'].items():
                final_grant_option = grant_option_override if grant_option_override is not None else grant_option
                desired_grants[obj][priv.upper()] = bool(final_grant_option)

        return desired_grants

    def update(self):
        desired = self._get_desired_grants()
        current = self.get()
        exclusive = self.module.params['exclusive']

        # Use set comprehensions for better performance
        all_current_privs = {(priv, obj, go)
                             for obj, privs in current.items()
                             for priv, go in privs.items()}

        all_desired_privs = {(priv, obj, go)
                             for obj, privs in desired.items()
                             for priv, go in privs.items()}

        to_revoke = all_current_privs - all_desired_privs if exclusive else set()
        to_grant = all_desired_privs - all_current_privs

        if not to_revoke and not to_grant:
            return self.changed

        self.changed = True

        queries = []
        revokes_by_obj = defaultdict(list)
        for priv, obj, go in to_revoke:
            revokes_by_obj[obj].append(priv)

        for obj, privs in revokes_by_obj.items():
            privs_str = ', '.join(sorted(privs))
            query = "REVOKE {0} ON {1} FROM '{2}'".format(privs_str, obj, self.grantee)
            if self.cluster:
                query += " ON CLUSTER {0}".format(self.cluster)
            queries.append(query)

        grants_go_by_obj = defaultdict(list)
        grants_no_go_by_obj = defaultdict(list)

        for priv, obj, go in to_grant:
            if go:
                grants_go_by_obj[obj].append(priv)
            else:
                grants_no_go_by_obj[obj].append(priv)

        for obj, privs in grants_go_by_obj.items():
            privs_str = ', '.join(sorted(privs))
            query = "GRANT {0} ON {1} TO '{2}' WITH GRANT OPTION".format(privs_str, obj, self.grantee)
            if self.cluster:
                query += " ON CLUSTER {0}".format(self.cluster)
            queries.append(query)

        for obj, privs in grants_no_go_by_obj.items():
            privs_str = ', '.join(sorted(privs))
            query = "GRANT {0} ON {1} TO '{2}'".format(privs_str, obj, self.grantee)
            if self.cluster:
                query += " ON CLUSTER {0}".format(self.cluster)
            queries.append(query)

        executed_statements.extend(queries)

        if self.module.check_mode:
            return self.changed

        for query in queries:
            execute_query(self.module, self.client, query)

        return self.changed

    def revoke(self):
        current = self.get()

        if not current:
            # No grants to revoke
            return self.changed

        self.changed = True

        # Build revoke queries for all current privileges
        queries = []
        for obj, privs in current.items():
            privs_str = ', '.join(sorted(privs))
            query = "REVOKE {0} ON {1} FROM '{2}'".format(privs_str, obj, self.grantee)
            if self.cluster:
                query += " ON CLUSTER {0}".format(self.cluster)
            queries.append(query)

        executed_statements.extend(queries)

        if self.module.check_mode:
            return self.changed

        for query in queries:
            execute_query(self.module, self.client, query)

        return self.changed


def main():
    argument_spec = client_common_argument_spec()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        grantee=dict(type='str', required=True),
        exclusive=dict(type='bool', default=False),
        privileges=dict(type='list', elements='dict'),
        cluster=dict(type='str', default=None),
    )

    # Instantiate an object of module class
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ('state', 'present', ['privileges']),
        ],
    )

    # Assign passed options to variables
    client_kwargs = module.params['client_kwargs']
    # The reason why these arguments are separate from client_kwargs
    # is that we need to protect some sensitive data like passwords passed
    # to the module from logging (see the arguments above with no_log=True);
    # Such data must be passed as module arguments (not nested deep in values).
    main_conn_kwargs = get_main_conn_kwargs(module)
    state = module.params['state']
    grantee = module.params['grantee']
    cluster = module.params['cluster']

    # Will fail if no driver informing the user
    check_clickhouse_driver(module)

    # Connect to DB
    client = connect_to_db_via_client(module, main_conn_kwargs, client_kwargs)

    # Do the job
    changed = False
    grants = ClickHouseGrants(module, client, grantee, cluster)
    # Get current grants
    start_grants = grants.get()

    if state == 'present':
        changed = grants.update()
    elif state == 'absent':
        changed = grants.revoke()

    # Get end grants - in check_mode, compute expected state
    if module.check_mode and changed:
        # In check mode, compute what the end state would be
        if state == 'absent':
            end_grants = {}
        else:  # state == 'present'
            desired_grants = grants._get_desired_grants()
            if module.params['exclusive']:
                end_grants = desired_grants
            else:
                # Merge current and desired grants (append mode)
                end_grants = dict(start_grants)
                for obj, privs in desired_grants.items():
                    if obj not in end_grants:
                        end_grants[obj] = {}
                    end_grants[obj].update(privs)
    else:
        # In normal mode or if no changes, query actual state
        end_grants = grants.get()

    # Close connection
    client.disconnect_connection()

    # Prepare result
    result = {
        'changed': changed,
        'executed_statements': executed_statements,
    }

    # Add diff if in diff mode or check mode
    if module._diff or module.check_mode:
        result['diff'] = {
            'before': start_grants,
            'after': end_grants,
        }

    # Users will get this in JSON output after execution
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2018, Dag Wieers (dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: cobbler_sync
short_description: Sync Cobbler
description:
  - Sync Cobbler to commit changes.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  host:
    description:
      - The name or IP address of the Cobbler system.
    default: 127.0.0.1
    type: str
  port:
    description:
      - Port number to be used for REST connection.
      - The default value depends on parameter O(use_ssl).
    type: int
  username:
    description:
      - The username to log in to Cobbler.
    default: cobbler
    type: str
  password:
    description:
      - The password to log in to Cobbler.
    type: str
  use_ssl:
    description:
      - If V(false), an HTTP connection is used instead of the default HTTPS connection.
    type: bool
    default: true
  validate_certs:
    description:
      - If V(false), SSL certificates are not validated.
      - This should only set to V(false) when used on personally controlled sites using self-signed certificates.
    type: bool
    default: true
author:
  - Dag Wieers (@dagwieers)
todo:
notes:
  - Concurrently syncing Cobbler is bound to fail with weird errors.
"""

EXAMPLES = r"""
- name: Commit Cobbler changes
  community.general.cobbler_sync:
    host: cobbler01
    username: cobbler
    password: MySuperSecureP4sswOrd
  run_once: true
  delegate_to: localhost
"""

RETURN = r"""
# Default return values
"""

import ssl
import xmlrpc.client as xmlrpc_client

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._datetime import (
    now,
)


def main():
    module = AnsibleModule(
        argument_spec=dict(
            host=dict(type="str", default="127.0.0.1"),
            port=dict(type="int"),
            username=dict(type="str", default="cobbler"),
            password=dict(type="str", no_log=True),
            use_ssl=dict(type="bool", default=True),
            validate_certs=dict(type="bool", default=True),
        ),
        supports_check_mode=True,
    )

    username = module.params["username"]
    password = module.params["password"]
    port = module.params["port"]
    use_ssl = module.params["use_ssl"]
    validate_certs = module.params["validate_certs"]

    proto = "https" if use_ssl else "http"
    if not port:
        port = "443" if use_ssl else "80"

    result = dict(
        changed=True,
    )

    start = now()

    ssl_context = None if validate_certs or not use_ssl else ssl._create_unverified_context()

    url = f"{proto}://{module.params['host']}:{port}/cobbler_api"
    conn = xmlrpc_client.ServerProxy(url, context=ssl_context)

    try:
        token = conn.login(username, password)
    except xmlrpc_client.Fault as e:
        module.fail_json(msg=f"Failed to log in to Cobbler '{url}' as '{username}'. {e}")
    except Exception as e:
        module.fail_json(msg=f"Connection to '{url}' failed. {e}")

    if not module.check_mode:
        try:
            conn.sync(token)
        except Exception as e:
            module.fail_json(msg=f"Failed to sync Cobbler. {e}")

    elapsed = now() - start
    module.exit_json(elapsed=elapsed.seconds, **result)


if __name__ == "__main__":
    main()



# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2026 Christoph Fiehe <christoph.fiehe@gmail.com>

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import typing as t

from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.urls import fetch_url, url_argument_spec

if t.TYPE_CHECKING:
    from http.client import HTTPResponse
    from urllib.error import HTTPError

    from ansible.module_utils.basic import AnsibleModule


class Icinga2Client:
    def __init__(
        self,
        module: AnsibleModule,
        url: str,
        ca_path: str | None = None,
        timeout: int | float | None = None,
    ) -> None:
        self.module = module
        self.url = url.rstrip("/")
        self.ca_path = ca_path
        self.timeout = timeout
        self.actions = Actions(client=self)

    def send_request(
        self, method: str, path: str, data: dict[str, t.Any] | None = None
    ) -> tuple[HTTPResponse | HTTPError, dict[str, t.Any]]:
        url = f"{self.url}/{path}"
        headers = {
            "X-HTTP-Method-Override": method.upper(),
            "Accept": "application/json",
        }
        return fetch_url(
            module=self.module,
            url=url,
            ca_path=self.ca_path,
            data=to_bytes(json.dumps(data)),
            headers=headers,
            timeout=self.timeout,
        )


class Actions:
    base_path = "v1/actions"

    def __init__(self, client: Icinga2Client) -> None:
        self.client = client

    def schedule_downtime(
        self,
        object_type: str,
        filter: str,
        author: str,
        comment: str,
        start_time: int,
        end_time: int,
        duration: int,
        filter_vars: dict[str, t.Any] | None = None,
        fixed: bool | None = None,
        all_services: bool | None = None,
        trigger_name: str | None = None,
        child_options: str | None = None,
    ) -> tuple[HTTPResponse | HTTPError, dict[str, t.Any]]:
        path = f"{self.base_path}/schedule-downtime"

        data: dict[str, t.Any] = {
            "type": object_type,
            "filter": filter,
            "author": author,
            "comment": comment,
            "start_time": start_time,
            "end_time": end_time,
            "duration": duration,
        }
        if filter_vars is not None:
            data["filter_vars"] = filter_vars
        if fixed is not None:
            data["fixed"] = fixed
        if all_services is not None:
            data["all_services"] = all_services
        if trigger_name is not None:
            data["trigger_name"] = trigger_name
        if child_options is not None:
            data["child_options"] = child_options

        return self.client.send_request(method="POST", path=path, data=data)

    def remove_downtime(
        self,
        object_type: str,
        name: str | None = None,
        filter: str | None = None,
        filter_vars: dict[str, t.Any] | None = None,
    ) -> tuple[HTTPResponse | HTTPError, dict[str, t.Any]]:
        path = f"{self.base_path}/remove-downtime"

        data: dict[str, t.Any] = {"type": object_type}
        if name is not None:
            data[object_type.lower()] = name
        if filter is not None:
            data["filter"] = filter
        if filter_vars is not None:
            data["filter_vars"] = filter_vars

        return self.client.send_request(method="POST", path=path, data=data)


def icinga2_argument_spec() -> dict[str, t.Any]:
    argument_spec = url_argument_spec()
    argument_spec.update(
        url=dict(type="str", required=True),
        ca_path=dict(type="path"),
        timeout=dict(type="int", default=10),
    )
    return argument_spec



# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:

    # Standard files documentation fragment

    # NOTE: This document fragment needs to be augmented by ZONE_ID_TYPE in a provider document fragment.
    #       The ZONE_ID_TYPE fragment will provide `choices` for the options.type entry.
    DOCUMENTATION = r"""
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  idempotent:
    support: full
options:
  state:
    description:
      - Specifies the state of the resource record.
    required: true
    choices: ['present', 'absent']
    type: str
  zone_name:
    description:
      - The DNS zone to modify.
      - Exactly one of O(zone_name) and O(zone_id) must be specified.
    type: str
    aliases:
      - zone
  zone_id:
    description:
      - The ID of the DNS zone to modify.
      - Exactly one of O(zone_name) and O(zone_id) must be specified.
    version_added: 0.2.0
  record:
    description:
      - The full DNS record to create or delete.
      - Exactly one of O(record) and O(prefix) must be specified.
    type: str
  prefix:
    description:
      - The prefix of the DNS record.
      - This is the part of O(record) before O(zone_name). For example, if the record to be modified is C(www.example.com)
        for the zone C(example.com), the prefix is V(www). If the record in this example would be C(example.com), the prefix
        would be V('') (empty string).
      - Exactly one of O(record) and O(prefix) must be specified.
    type: str
    version_added: 0.2.0
  ttl:
    description:
      - The TTL to give the new record, in seconds.
    type: int
  type:
    description:
      - The type of DNS record to create or delete.
    required: true
    type: str
  value:
    description:
      - The new value when creating a DNS record.
      - YAML lists or multiple comma-spaced values are allowed.
      - When deleting a record all values for the record must be specified or it will not be deleted.
      - Must be specified if O(state=present) or when O(on_existing) is not V(replace).
      - Will be ignored if O(state=absent) and O(on_existing=replace).
    type: list
    elements: str
  on_existing:
    description:
      - This option defines the behavior if the record set already exists, but differs from the specified record set. For
        this comparison, O(value) and O(ttl) are used for all records of type O(type) matching the O(prefix) resp. O(record).
      - If set to V(replace), the record will be updated (O(state=present)) or removed (O(state=absent)). This is the old
        O(ignore:overwrite=true) behavior.
      - If set to V(keep_and_fail), the module will fail and not modify the records. This is the old O(ignore:overwrite=false)
        behavior if O(state=present).
      - If set to V(keep_and_warn), the module will warn and not modify the records.
      - If set to V(keep), the module will not modify the records. This is the old O(ignore:overwrite=false) behavior if O(state=absent).
      - If O(state=absent) and the value is not V(replace), O(value) must be specified.
    default: replace
    type: str
    choices:
      - replace
      - keep_and_fail
      - keep_and_warn
      - keep
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_camera_wireless_profiles_info
short_description: Information module for networks _camera _wireless _profiles
description:
  - Get all networks _camera _wireless _profiles.
  - Get networks _camera _wireless _profiles by id.
  - List the camera wireless profiles for this network.
  - Retrieve a single camera wireless profile.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  networkId:
    description:
      - NetworkId path parameter. Network ID.
    type: str
  wirelessProfileId:
    description:
      - WirelessProfileId path parameter. Wireless profile ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for camera getNetworkCameraWirelessProfile
    description: Complete reference of the getNetworkCameraWirelessProfile API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-network-camera-wireless-profile
  - name: Cisco Meraki documentation for camera getNetworkCameraWirelessProfiles
    description: Complete reference of the getNetworkCameraWirelessProfiles API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-network-camera-wireless-profiles
notes:
  - SDK Method used are
    camera.Camera.get_network_camera_wireless_profile,
    camera.Camera.get_network_camera_wireless_profiles,
  - Paths used are
    get /networks/{networkId}/camera/wirelessProfiles,
    get /networks/{networkId}/camera/wirelessProfiles/{wirelessProfileId},
"""

EXAMPLES = r"""
- name: Get all networks _camera _wireless _profiles
  cisco.meraki.networks_camera_wireless_profiles_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkId: string
  register: result
- name: Get networks _camera _wireless _profiles by id
  cisco.meraki.networks_camera_wireless_profiles_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    networkId: string
    wirelessProfileId: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "id": "string",
      "name": "string",
      "appliedDeviceCount": 0,
      "ssid": {
        "name": "string",
        "authMode": "string",
        "encryptionMode": "string",
        "psk": "string"
      },
      "identity": {
        "username": "string",
        "password": "string"
      }
    }
"""



#!/usr/bin/python
#
# Copyright (C) 2019 Huawei
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

###############################################################################
# Documentation
###############################################################################

DOCUMENTATION = r"""
module: hwc_vpc_port
description:
  - VPC port management.
short_description: Creates a resource of VPC/Port in Huawei Cloud
version_added: '0.2.0'
author: Huawei Inc. (@huaweicloud)
requirements:
  - keystoneauth1 >= 3.6.0
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  state:
    description:
      - Whether the given object should exist in Huawei Cloud.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  timeouts:
    description:
      - The timeouts for each operations.
    type: dict
    default: {}
    suboptions:
      create:
        description:
          - The timeouts for create operation.
        type: str
        default: '15m'
  subnet_id:
    description:
      - Specifies the ID of the subnet to which the port belongs.
    type: str
    required: true
  admin_state_up:
    description:
      - Specifies the administrative state of the port.
    type: bool
  allowed_address_pairs:
    description:
      - Specifies a set of zero or more allowed address pairs.
    type: list
    elements: dict
    suboptions:
      ip_address:
        description:
          - Specifies the IP address. It cannot set it to 0.0.0.0. Configure an independent security group for the port if
            a large CIDR block (subnet mask less than 24) is configured for parameter allowed_address_pairs.
        type: str
      mac_address:
        description:
          - Specifies the MAC address.
        type: str
  extra_dhcp_opts:
    description:
      - Specifies the extended option of DHCP.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - Specifies the option name.
        type: str
      value:
        description:
          - Specifies the option value.
        type: str
  ip_address:
    description:
      - Specifies the port IP address.
    type: str
  name:
    description:
      - Specifies the port name. The value can contain no more than 255 characters.
    type: str
  security_groups:
    description:
      - Specifies the ID of the security group.
    type: list
    elements: str
extends_documentation_fragment:
  - community.general._hwc
  - community.general._attributes
"""

EXAMPLES = r"""
# create a port
- name: Create vpc
  hwc_network_vpc:
    cidr: "192.168.100.0/24"
    name: "ansible_network_vpc_test"
  register: vpc
- name: Create subnet
  hwc_vpc_subnet:
    gateway_ip: "192.168.100.32"
    name: "ansible_network_subnet_test"
    dhcp_enable: true
    vpc_id: "{{ vpc.id }}"
    cidr: "192.168.100.0/26"
  register: subnet
- name: Create a port
  community.general.hwc_vpc_port:
    subnet_id: "{{ subnet.id }}"
    ip_address: "192.168.100.33"
"""

RETURN = r"""
subnet_id:
  description:
    - Specifies the ID of the subnet to which the port belongs.
  type: str
  returned: success
admin_state_up:
  description:
    - Specifies the administrative state of the port.
  type: bool
  returned: success
allowed_address_pairs:
  description:
    - Specifies a set of zero or more allowed address pairs.
  type: list
  returned: success
  contains:
    ip_address:
      description:
        - Specifies the IP address. It cannot set it to 0.0.0.0. Configure an independent security group for the port if a
          large CIDR block (subnet mask less than 24) is configured for parameter allowed_address_pairs.
      type: str
      returned: success
    mac_address:
      description:
        - Specifies the MAC address.
      type: str
      returned: success
extra_dhcp_opts:
  description:
    - Specifies the extended option of DHCP.
  type: list
  returned: success
  contains:
    name:
      description:
        - Specifies the option name.
      type: str
      returned: success
    value:
      description:
        - Specifies the option value.
      type: str
      returned: success
ip_address:
  description:
    - Specifies the port IP address.
  type: str
  returned: success
name:
  description:
    - Specifies the port name. The value can contain no more than 255 characters.
  type: str
  returned: success
security_groups:
  description:
    - Specifies the ID of the security group.
  type: list
  returned: success
mac_address:
  description:
    - Specifies the port MAC address.
  type: str
  returned: success
"""

from ansible_collections.community.general.plugins.module_utils._hwc_utils import (
    Config,
    HwcClientException,
    HwcClientException404,
    HwcModule,
    are_different_dicts,
    build_path,
    get_region,
    is_empty_value,
    navigate_value,
    wait_to_finish,
)


def build_module():
    return HwcModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "absent"], type="str"),
            timeouts=dict(
                type="dict",
                options=dict(
                    create=dict(default="15m", type="str"),
                ),
                default=dict(),
            ),
            subnet_id=dict(type="str", required=True),
            admin_state_up=dict(type="bool"),
            allowed_address_pairs=dict(
                type="list",
                elements="dict",
                options=dict(ip_address=dict(type="str"), mac_address=dict(type="str")),
            ),
            extra_dhcp_opts=dict(
                type="list", elements="dict", options=dict(name=dict(type="str"), value=dict(type="str"))
            ),
            ip_address=dict(type="str"),
            name=dict(type="str"),
            security_groups=dict(type="list", elements="str"),
        ),
        supports_check_mode=True,
    )


def main():
    """Main function"""

    module = build_module()
    config = Config(module, "vpc")

    try:
        resource = None
        if module.params["id"]:
            resource = True
        else:
            v = search_resource(config)
            if len(v) > 1:
                raise Exception(f"Found more than one resource({', '.join([navigate_value(i, ['id']) for i in v])})")

            if len(v) == 1:
                resource = v[0]
                module.params["id"] = navigate_value(resource, ["id"])

        result = {}
        changed = False
        if module.params["state"] == "present":
            if resource is None:
                if not module.check_mode:
                    create(config)
                changed = True

            current = read_resource(config, exclude_output=True)
            expect = user_input_parameters(module)
            if are_different_dicts(expect, current):
                if not module.check_mode:
                    update(config)
                changed = True

            result = read_resource(config)
            result["id"] = module.params.get("id")
        else:
            if resource:
                if not module.check_mode:
                    delete(config)
                changed = True

    except Exception as ex:
        module.fail_json(msg=str(ex))

    else:
        result["changed"] = changed
        module.exit_json(**result)


def user_input_parameters(module):
    return {
        "admin_state_up": module.params.get("admin_state_up"),
        "allowed_address_pairs": module.params.get("allowed_address_pairs"),
        "extra_dhcp_opts": module.params.get("extra_dhcp_opts"),
        "ip_address": module.params.get("ip_address"),
        "name": module.params.get("name"),
        "security_groups": module.params.get("security_groups"),
        "subnet_id": module.params.get("subnet_id"),
    }


def create(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    timeout = 60 * int(module.params["timeouts"]["create"].rstrip("m"))
    opts = user_input_parameters(module)

    params = build_create_parameters(opts)
    r = send_create_request(module, params, client)
    obj = async_wait_create(config, r, client, timeout)
    module.params["id"] = navigate_value(obj, ["port", "id"])


def update(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    opts = user_input_parameters(module)

    params = build_update_parameters(opts)
    if params:
        send_update_request(module, params, client)


def delete(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    send_delete_request(module, None, client)

    url = build_path(module, "ports/{id}")

    def _refresh_status():
        try:
            client.get(url)
        except HwcClientException404:
            return True, "Done"

        except Exception:
            return None, ""

        return True, "Pending"

    timeout = 60 * int(module.params["timeouts"]["create"].rstrip("m"))
    try:
        wait_to_finish(["Done"], ["Pending"], _refresh_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_vpc_port): error waiting for api(delete) to be done, error= {ex}")


def read_resource(config, exclude_output=False):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    res = {}

    r = send_read_request(module, client)
    res["read"] = fill_read_resp_body(r)

    array_index = {
        "read.fixed_ips": 0,
    }

    return update_properties(module, res, array_index, exclude_output)


def _build_query_link(opts):
    query_params = []

    v = navigate_value(opts, ["subnet_id"])
    if v:
        query_params.append(f"network_id={v}")

    v = navigate_value(opts, ["name"])
    if v:
        query_params.append(f"name={v}")

    v = navigate_value(opts, ["admin_state_up"])
    if v:
        query_params.append(f"admin_state_up={v}")

    query_link = "?marker={marker}&limit=10"
    if query_params:
        query_link += f"&{'&'.join(query_params)}"

    return query_link


def search_resource(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    opts = user_input_parameters(module)
    identity_obj = _build_identity_object(opts)
    query_link = _build_query_link(opts)
    link = f"ports{query_link}"

    result = []
    p = {"marker": ""}
    while True:
        url = link.format(**p)
        r = send_list_request(module, client, url)
        if not r:
            break

        for item in r:
            item = fill_list_resp_body(item)
            if not are_different_dicts(identity_obj, item):
                result.append(item)

        if len(result) > 1:
            break

        p["marker"] = r[-1].get("id")

    return result


def build_create_parameters(opts):
    params = dict()

    v = navigate_value(opts, ["admin_state_up"], None)
    if not is_empty_value(v):
        params["admin_state_up"] = v

    v = expand_create_allowed_address_pairs(opts, None)
    if not is_empty_value(v):
        params["allowed_address_pairs"] = v

    v = expand_create_extra_dhcp_opts(opts, None)
    if not is_empty_value(v):
        params["extra_dhcp_opts"] = v

    v = expand_create_fixed_ips(opts, None)
    if not is_empty_value(v):
        params["fixed_ips"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    v = navigate_value(opts, ["subnet_id"], None)
    if not is_empty_value(v):
        params["network_id"] = v

    v = navigate_value(opts, ["security_groups"], None)
    if not is_empty_value(v):
        params["security_groups"] = v

    if not params:
        return params

    params = {"port": params}

    return params


def expand_create_allowed_address_pairs(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    v = navigate_value(d, ["allowed_address_pairs"], new_array_index)
    if not v:
        return req
    n = len(v)
    for i in range(n):
        new_array_index["allowed_address_pairs"] = i
        transformed = dict()

        v = navigate_value(d, ["allowed_address_pairs", "ip_address"], new_array_index)
        if not is_empty_value(v):
            transformed["ip_address"] = v

        v = navigate_value(d, ["allowed_address_pairs", "mac_address"], new_array_index)
        if not is_empty_value(v):
            transformed["mac_address"] = v

        if transformed:
            req.append(transformed)

    return req


def expand_create_extra_dhcp_opts(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    v = navigate_value(d, ["extra_dhcp_opts"], new_array_index)
    if not v:
        return req
    n = len(v)
    for i in range(n):
        new_array_index["extra_dhcp_opts"] = i
        transformed = dict()

        v = navigate_value(d, ["extra_dhcp_opts", "name"], new_array_index)
        if not is_empty_value(v):
            transformed["opt_name"] = v

        v = navigate_value(d, ["extra_dhcp_opts", "value"], new_array_index)
        if not is_empty_value(v):
            transformed["opt_value"] = v

        if transformed:
            req.append(transformed)

    return req


def expand_create_fixed_ips(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    transformed = dict()

    v = navigate_value(d, ["ip_address"], new_array_index)
    if not is_empty_value(v):
        transformed["ip_address"] = v

    if transformed:
        req.append(transformed)

    return req


def send_create_request(module, params, client):
    url = "ports"
    try:
        r = client.post(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_port): error running api(create), error: {ex}"
        module.fail_json(msg=msg)

    return r


def async_wait_create(config, result, client, timeout):
    module = config.module

    path_parameters = {
        "port_id": ["port", "id"],
    }
    data = {key: navigate_value(result, path) for key, path in path_parameters.items()}

    url = build_path(module, "ports/{port_id}", data)

    def _query_status():
        r = None
        try:
            r = client.get(url, timeout=timeout)
        except HwcClientException:
            return None, ""

        try:
            s = navigate_value(r, ["port", "status"])
            return r, s
        except Exception:
            return None, ""

    try:
        return wait_to_finish(["ACTIVE", "DOWN"], ["BUILD"], _query_status, timeout)
    except Exception as ex:
        module.fail_json(msg=f"module(hwc_vpc_port): error waiting for api(create) to be done, error= {ex}")


def build_update_parameters(opts):
    params = dict()

    v = expand_update_allowed_address_pairs(opts, None)
    if v is not None:
        params["allowed_address_pairs"] = v

    v = expand_update_extra_dhcp_opts(opts, None)
    if v is not None:
        params["extra_dhcp_opts"] = v

    v = navigate_value(opts, ["name"], None)
    if not is_empty_value(v):
        params["name"] = v

    v = navigate_value(opts, ["security_groups"], None)
    if not is_empty_value(v):
        params["security_groups"] = v

    if not params:
        return params

    params = {"port": params}

    return params


def expand_update_allowed_address_pairs(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    v = navigate_value(d, ["allowed_address_pairs"], new_array_index)
    if not v:
        return req
    n = len(v)
    for i in range(n):
        new_array_index["allowed_address_pairs"] = i
        transformed = dict()

        v = navigate_value(d, ["allowed_address_pairs", "ip_address"], new_array_index)
        if not is_empty_value(v):
            transformed["ip_address"] = v

        v = navigate_value(d, ["allowed_address_pairs", "mac_address"], new_array_index)
        if not is_empty_value(v):
            transformed["mac_address"] = v

        if transformed:
            req.append(transformed)

    return req


def expand_update_extra_dhcp_opts(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    v = navigate_value(d, ["extra_dhcp_opts"], new_array_index)
    if not v:
        return req
    n = len(v)
    for i in range(n):
        new_array_index["extra_dhcp_opts"] = i
        transformed = dict()

        v = navigate_value(d, ["extra_dhcp_opts", "name"], new_array_index)
        if not is_empty_value(v):
            transformed["opt_name"] = v

        v = navigate_value(d, ["extra_dhcp_opts", "value"], new_array_index)
        if not is_empty_value(v):
            transformed["opt_value"] = v

        if transformed:
            req.append(transformed)

    return req


def send_update_request(module, params, client):
    url = build_path(module, "ports/{id}")

    try:
        r = client.put(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_port): error running api(update), error: {ex}"
        module.fail_json(msg=msg)

    return r


def send_delete_request(module, params, client):
    url = build_path(module, "ports/{id}")

    try:
        r = client.delete(url, params)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_port): error running api(delete), error: {ex}"
        module.fail_json(msg=msg)

    return r


def send_read_request(module, client):
    url = build_path(module, "ports/{id}")

    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_port): error running api(read), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["port"], None)


def fill_read_resp_body(body):
    result = dict()

    result["admin_state_up"] = body.get("admin_state_up")

    v = fill_read_resp_allowed_address_pairs(body.get("allowed_address_pairs"))
    result["allowed_address_pairs"] = v

    result["binding_host_id"] = body.get("binding_host_id")

    result["binding_vnic_type"] = body.get("binding_vnic_type")

    result["device_id"] = body.get("device_id")

    result["device_owner"] = body.get("device_owner")

    result["dns_name"] = body.get("dns_name")

    v = fill_read_resp_extra_dhcp_opts(body.get("extra_dhcp_opts"))
    result["extra_dhcp_opts"] = v

    v = fill_read_resp_fixed_ips(body.get("fixed_ips"))
    result["fixed_ips"] = v

    result["id"] = body.get("id")

    result["mac_address"] = body.get("mac_address")

    result["name"] = body.get("name")

    result["network_id"] = body.get("network_id")

    result["security_groups"] = body.get("security_groups")

    result["status"] = body.get("status")

    result["tenant_id"] = body.get("tenant_id")

    return result


def fill_read_resp_allowed_address_pairs(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["ip_address"] = item.get("ip_address")

        val["mac_address"] = item.get("mac_address")

        result.append(val)

    return result


def fill_read_resp_extra_dhcp_opts(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["opt_name"] = item.get("opt_name")

        val["opt_value"] = item.get("opt_value")

        result.append(val)

    return result


def fill_read_resp_fixed_ips(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["ip_address"] = item.get("ip_address")

        result.append(val)

    return result


def update_properties(module, response, array_index, exclude_output=False):
    r = user_input_parameters(module)

    v = navigate_value(response, ["read", "admin_state_up"], array_index)
    r["admin_state_up"] = v

    v = r.get("allowed_address_pairs")
    v = flatten_allowed_address_pairs(response, array_index, v, exclude_output)
    r["allowed_address_pairs"] = v

    v = r.get("extra_dhcp_opts")
    v = flatten_extra_dhcp_opts(response, array_index, v, exclude_output)
    r["extra_dhcp_opts"] = v

    v = navigate_value(response, ["read", "fixed_ips", "ip_address"], array_index)
    r["ip_address"] = v

    if not exclude_output:
        v = navigate_value(response, ["read", "mac_address"], array_index)
        r["mac_address"] = v

    v = navigate_value(response, ["read", "name"], array_index)
    r["name"] = v

    v = navigate_value(response, ["read", "security_groups"], array_index)
    r["security_groups"] = v

    v = navigate_value(response, ["read", "network_id"], array_index)
    r["subnet_id"] = v

    return r


def flatten_allowed_address_pairs(d, array_index, current_value, exclude_output):
    n = 0
    result = current_value
    has_init_value = True
    if result:
        n = len(result)
    else:
        has_init_value = False
        result = []
        v = navigate_value(d, ["read", "allowed_address_pairs"], array_index)
        if not v:
            return current_value
        n = len(v)

    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    for i in range(n):
        new_array_index["read.allowed_address_pairs"] = i

        val = dict()
        if len(result) >= (i + 1) and result[i]:
            val = result[i]

        v = navigate_value(d, ["read", "allowed_address_pairs", "ip_address"], new_array_index)
        val["ip_address"] = v

        v = navigate_value(d, ["read", "allowed_address_pairs", "mac_address"], new_array_index)
        val["mac_address"] = v

        if len(result) >= (i + 1):
            result[i] = val
        else:
            for v in val.values():
                if v is not None:
                    result.append(val)
                    break

    return result if (has_init_value or result) else current_value


def flatten_extra_dhcp_opts(d, array_index, current_value, exclude_output):
    n = 0
    result = current_value
    has_init_value = True
    if result:
        n = len(result)
    else:
        has_init_value = False
        result = []
        v = navigate_value(d, ["read", "extra_dhcp_opts"], array_index)
        if not v:
            return current_value
        n = len(v)

    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    for i in range(n):
        new_array_index["read.extra_dhcp_opts"] = i

        val = dict()
        if len(result) >= (i + 1) and result[i]:
            val = result[i]

        v = navigate_value(d, ["read", "extra_dhcp_opts", "opt_name"], new_array_index)
        val["name"] = v

        v = navigate_value(d, ["read", "extra_dhcp_opts", "opt_value"], new_array_index)
        val["value"] = v

        if len(result) >= (i + 1):
            result[i] = val
        else:
            for v in val.values():
                if v is not None:
                    result.append(val)
                    break

    return result if (has_init_value or result) else current_value


def send_list_request(module, client, url):
    r = None
    try:
        r = client.get(url)
    except HwcClientException as ex:
        msg = f"module(hwc_vpc_port): error running api(list), error: {ex}"
        module.fail_json(msg=msg)

    return navigate_value(r, ["ports"], None)


def _build_identity_object(all_opts):
    result = dict()

    v = navigate_value(all_opts, ["admin_state_up"], None)
    result["admin_state_up"] = v

    v = expand_list_allowed_address_pairs(all_opts, None)
    result["allowed_address_pairs"] = v

    result["binding_host_id"] = None

    result["binding_vnic_type"] = None

    result["device_id"] = None

    result["device_owner"] = None

    result["dns_name"] = None

    v = expand_list_extra_dhcp_opts(all_opts, None)
    result["extra_dhcp_opts"] = v

    v = expand_list_fixed_ips(all_opts, None)
    result["fixed_ips"] = v

    result["id"] = None

    result["mac_address"] = None

    v = navigate_value(all_opts, ["name"], None)
    result["name"] = v

    v = navigate_value(all_opts, ["subnet_id"], None)
    result["network_id"] = v

    v = navigate_value(all_opts, ["security_groups"], None)
    result["security_groups"] = v

    result["status"] = None

    result["tenant_id"] = None

    return result


def expand_list_allowed_address_pairs(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    v = navigate_value(d, ["allowed_address_pairs"], new_array_index)

    n = len(v) if v else 1
    for i in range(n):
        new_array_index["allowed_address_pairs"] = i
        transformed = dict()

        v = navigate_value(d, ["allowed_address_pairs", "ip_address"], new_array_index)
        transformed["ip_address"] = v

        v = navigate_value(d, ["allowed_address_pairs", "mac_address"], new_array_index)
        transformed["mac_address"] = v

        for v in transformed.values():
            if v is not None:
                req.append(transformed)
                break

    return req if req else None


def expand_list_extra_dhcp_opts(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    v = navigate_value(d, ["extra_dhcp_opts"], new_array_index)

    n = len(v) if v else 1
    for i in range(n):
        new_array_index["extra_dhcp_opts"] = i
        transformed = dict()

        v = navigate_value(d, ["extra_dhcp_opts", "name"], new_array_index)
        transformed["opt_name"] = v

        v = navigate_value(d, ["extra_dhcp_opts", "value"], new_array_index)
        transformed["opt_value"] = v

        for v in transformed.values():
            if v is not None:
                req.append(transformed)
                break

    return req if req else None


def expand_list_fixed_ips(d, array_index):
    new_array_index = dict()
    if array_index:
        new_array_index.update(array_index)

    req = []

    transformed = dict()

    v = navigate_value(d, ["ip_address"], new_array_index)
    transformed["ip_address"] = v

    for v in transformed.values():
        if v is not None:
            req.append(transformed)
            break

    return req if req else None


def fill_list_resp_body(body):
    result = dict()

    result["admin_state_up"] = body.get("admin_state_up")

    v = fill_list_resp_allowed_address_pairs(body.get("allowed_address_pairs"))
    result["allowed_address_pairs"] = v

    result["binding_host_id"] = body.get("binding_host_id")

    result["binding_vnic_type"] = body.get("binding_vnic_type")

    result["device_id"] = body.get("device_id")

    result["device_owner"] = body.get("device_owner")

    result["dns_name"] = body.get("dns_name")

    v = fill_list_resp_extra_dhcp_opts(body.get("extra_dhcp_opts"))
    result["extra_dhcp_opts"] = v

    v = fill_list_resp_fixed_ips(body.get("fixed_ips"))
    result["fixed_ips"] = v

    result["id"] = body.get("id")

    result["mac_address"] = body.get("mac_address")

    result["name"] = body.get("name")

    result["network_id"] = body.get("network_id")

    result["security_groups"] = body.get("security_groups")

    result["status"] = body.get("status")

    result["tenant_id"] = body.get("tenant_id")

    return result


def fill_list_resp_allowed_address_pairs(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["ip_address"] = item.get("ip_address")

        val["mac_address"] = item.get("mac_address")

        result.append(val)

    return result


def fill_list_resp_extra_dhcp_opts(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["opt_name"] = item.get("opt_name")

        val["opt_value"] = item.get("opt_value")

        result.append(val)

    return result


def fill_list_resp_fixed_ips(value):
    if not value:
        return None

    result = []
    for item in value:
        val = dict()

        val["ip_address"] = item.get("ip_address")

        result.append(val)

    return result


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The ios_lacp_interfaces fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type


import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.lacp_interfaces.lacp_interfaces import (
    Lacp_InterfacesArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.utils.utils import (
    get_interface_type,
    normalize_interface,
)


class Lacp_InterfacesFacts(object):
    """The ios_lacp_interfaces fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Lacp_InterfacesArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_lacp_interface_data(self, connection):
        return connection.get("show running-config | section ^interface")

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for lacp_interfaces
        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """

        objs = []
        if not data:
            data = self.get_lacp_interface_data(connection)

        # operate on a collection of resource x
        config = ("\n" + data).split("\ninterface ")

        for conf in config:
            if conf:
                obj = self.render_config(self.generated_spec, conf)
                if obj:
                    objs.append(obj)
        facts = {}

        if objs:
            facts["lacp_interfaces"] = []
            params = utils.validate_config(self.argument_spec, {"config": objs})
            for cfg in params["config"]:
                facts["lacp_interfaces"].append(utils.remove_empties(cfg))
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts

    def render_config(self, spec, conf):
        """
        Render config as dictionary structure and delete keys
          from spec for null values

        :param spec: The facts tree, generated from the argspec
        :param conf: The configuration
        :rtype: dictionary
        :returns: The generated config
        """
        config = deepcopy(spec)
        match = re.search(r"^(\S+)", conf)
        intf = match.group(1)
        if get_interface_type(intf) == "unknown":
            return {}

        config["name"] = normalize_interface(intf)
        port_priority = utils.parse_conf_arg(conf, "lacp port-priority")
        max_bundle = utils.parse_conf_arg(conf, "lacp max-bundle")
        if port_priority:
            config["port_priority"] = int(port_priority)
        if "lacp fast-switchover" in conf:
            config["fast_switchover"] = True
        if max_bundle:
            config["max_bundle"] = int(max_bundle)

        return utils.remove_empties(config)



# Copyright (c) 2016, Yanis Guenane <yanis+ansible@guenane.org>
# Copyright (c) 2020, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

try:
    import cryptography  # noqa: F401, pylint: disable=unused-import

    HAS_CRYPTOGRAPHY = True
except ImportError:
    # Error handled in the calling module.
    HAS_CRYPTOGRAPHY = False


class OpenSSLObjectError(Exception):
    pass


class OpenSSLBadPassphraseError(OpenSSLObjectError):
    pass


__all__ = ("HAS_CRYPTOGRAPHY", "OpenSSLBadPassphraseError", "OpenSSLObjectError")



# Copyright (c) 2021 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
name: xen_orchestra
short_description: Xen Orchestra inventory source
version_added: 4.1.0
author:
  - Dom Del Nano (@ddelnano) <ddelnano@gmail.com>
  - Samori Gorse (@shinuza) <samorigorse@gmail.com>
requirements:
  - websocket-client >= 1.0.0
description:
  - Get inventory hosts from a Xen Orchestra deployment.
  - Uses a configuration file as an inventory source, it must end in C(.xen_orchestra.yml) or C(.xen_orchestra.yaml).
extends_documentation_fragment:
  - ansible.builtin.constructed
  - ansible.builtin.inventory_cache
options:
  plugin:
    description: The name of this plugin, it should always be set to V(community.general.xen_orchestra) for this plugin to
      recognize it as its own.
    required: true
    choices: ['community.general.xen_orchestra']
    type: str
  api_host:
    description:
      - API host to XOA API.
      - If the value is not specified in the inventory configuration, the value of environment variable E(ANSIBLE_XO_HOST)
        is used instead.
    type: str
    env:
      - name: ANSIBLE_XO_HOST
  user:
    description:
      - Xen Orchestra user.
      - If the value is not specified in the inventory configuration, the value of environment variable E(ANSIBLE_XO_USER)
        is used instead.
    required: true
    type: str
    env:
      - name: ANSIBLE_XO_USER
  password:
    description:
      - Xen Orchestra password.
      - If the value is not specified in the inventory configuration, the value of environment variable E(ANSIBLE_XO_PASSWORD)
        is used instead.
    required: true
    type: str
    env:
      - name: ANSIBLE_XO_PASSWORD
  validate_certs:
    description: Verify TLS certificate if using HTTPS.
    type: boolean
    default: true
  use_ssl:
    description: Use wss when connecting to the Xen Orchestra API.
    type: boolean
    default: true
  use_vm_uuid:
    description:
      - Import Xen VMs to inventory using their UUID as the VM entry name.
      - If set to V(false) use VM name labels instead of UUIDs.
    type: boolean
    default: true
    version_added: 10.4.0
  use_host_uuid:
    description:
      - Import Xen Hosts to inventory using their UUID as the Host entry name.
      - If set to V(false) use Host name labels instead of UUIDs.
    type: boolean
    default: true
    version_added: 10.4.0
"""


EXAMPLES = r"""
---
# file must be named xen_orchestra.yaml or xen_orchestra.yml
plugin: community.general.xen_orchestra
api_host: 192.168.1.255
user: xo
password: xo_pwd
validate_certs: true
use_ssl: true
groups:
  kube_nodes: "'kube_node' in tags"
compose:
  ansible_port: 2222
use_vm_uuid: false
use_host_uuid: true
"""

import json
import ssl
from time import sleep

from ansible.errors import AnsibleError
from ansible.plugins.inventory import BaseInventoryPlugin, Cacheable, Constructable

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion
from ansible_collections.community.general.plugins.plugin_utils._unsafe import make_unsafe

# 3rd party imports
try:
    HAS_WEBSOCKET = True
    import websocket
    from websocket import create_connection

    if LooseVersion(websocket.__version__) <= LooseVersion("1.0.0"):
        raise ImportError
except ImportError:
    HAS_WEBSOCKET = False


HALTED = "Halted"
PAUSED = "Paused"
RUNNING = "Running"
SUSPENDED = "Suspended"
POWER_STATES = [RUNNING, HALTED, SUSPENDED, PAUSED]
HOST_GROUP = "xo_hosts"
POOL_GROUP = "xo_pools"


def clean_group_name(label):
    return label.lower().replace(" ", "-").replace("-", "_")


class InventoryModule(BaseInventoryPlugin, Constructable, Cacheable):
    """Host inventory parser for ansible using XenOrchestra as source."""

    NAME = "community.general.xen_orchestra"

    def __init__(self):
        super().__init__()

        # from config
        self.counter = -1
        self.session = None
        self.cache_key = None
        self.use_cache = None

    @property
    def pointer(self):
        self.counter += 1
        return self.counter

    def create_connection(self, xoa_api_host):
        validate_certs = self.get_option("validate_certs")
        use_ssl = self.get_option("use_ssl")
        proto = "wss" if use_ssl else "ws"

        sslopt = None if validate_certs else {"cert_reqs": ssl.CERT_NONE}
        self.conn = create_connection(f"{proto}://{xoa_api_host}/api/", sslopt=sslopt)

    CALL_TIMEOUT = 100
    """Number of 1/10ths of a second to wait before method call times out."""

    def call(self, method, params):
        """Calls a method on the XO server with the provided parameters."""
        id = self.pointer
        self.conn.send(json.dumps({"id": id, "jsonrpc": "2.0", "method": method, "params": params}))

        waited = 0
        while waited < self.CALL_TIMEOUT:
            response = json.loads(self.conn.recv())
            if "id" in response and response["id"] == id:
                return response
            else:
                sleep(0.1)
                waited += 1

        raise AnsibleError(f"Method call {method} timed out after {self.CALL_TIMEOUT / 10} seconds.")

    def login(self, user, password):
        result = self.call("session.signIn", {"username": user, "password": password})

        if "error" in result:
            raise AnsibleError(f"Could not connect: {result['error']}")

    def get_object(self, name):
        answer = self.call("xo.getAllObjects", {"filter": {"type": name}})

        if "error" in answer:
            raise AnsibleError(f"Could not request: {answer['error']}")

        return answer["result"]

    def _get_objects(self):
        self.create_connection(self.xoa_api_host)
        self.login(self.xoa_user, self.xoa_password)

        return {
            "vms": self.get_object("VM"),
            "pools": self.get_object("pool"),
            "hosts": self.get_object("host"),
        }

    def _apply_constructable(self, name, variables):
        strict = self.get_option("strict")
        self._add_host_to_composed_groups(self.get_option("groups"), variables, name, strict=strict)
        self._add_host_to_keyed_groups(self.get_option("keyed_groups"), variables, name, strict=strict)
        self._set_composite_vars(self.get_option("compose"), variables, name, strict=strict)

    def _add_vms(self, vms, hosts, pools):
        vm_name_list = []
        for uuid, vm in vms.items():
            if self.vm_entry_name_type == "name_label":
                if vm["name_label"] not in vm_name_list:
                    entry_name = vm["name_label"]
                    vm_name_list.append(vm["name_label"])
                else:
                    vm_duplicate_count = vm_name_list.count(vm["name_label"])
                    entry_name = f"{vm['name_label']}_{vm_duplicate_count}"
                    vm_name_list.append(vm["name_label"])
            else:
                entry_name = uuid
            group = "with_ip"
            ip = vm.get("mainIpAddress")
            power_state = vm["power_state"].lower()
            pool_name = self._pool_group_name_for_uuid(pools, vm["$poolId"])
            host_name = self._host_group_name_for_uuid(hosts, vm["$container"])

            self.inventory.add_host(entry_name)

            # Grouping by power state
            self.inventory.add_child(power_state, entry_name)

            # Grouping by host
            if host_name:
                self.inventory.add_child(host_name, entry_name)

            # Grouping by pool
            if pool_name:
                self.inventory.add_child(pool_name, entry_name)

            # Grouping VMs with an IP together
            if ip is None:
                group = "without_ip"
            self.inventory.add_group(group)
            self.inventory.add_child(group, entry_name)

            # Adding meta
            self.inventory.set_variable(entry_name, "uuid", uuid)
            self.inventory.set_variable(entry_name, "ip", ip)
            self.inventory.set_variable(entry_name, "ansible_host", ip)
            self.inventory.set_variable(entry_name, "power_state", power_state)
            self.inventory.set_variable(entry_name, "name_label", vm["name_label"])
            self.inventory.set_variable(entry_name, "type", vm["type"])
            self.inventory.set_variable(entry_name, "cpus", vm["CPUs"]["number"])
            self.inventory.set_variable(entry_name, "tags", vm["tags"])
            self.inventory.set_variable(entry_name, "memory", vm["memory"]["size"])
            self.inventory.set_variable(entry_name, "has_ip", group == "with_ip")
            self.inventory.set_variable(entry_name, "is_managed", vm.get("managementAgentDetected", False))
            self.inventory.set_variable(entry_name, "os_version", vm["os_version"])

            self._apply_constructable(entry_name, self.inventory.get_host(entry_name).get_vars())

    def _add_hosts(self, hosts, pools):
        host_name_list = []
        for host in hosts.values():
            if self.host_entry_name_type == "name_label":
                if host["name_label"] not in host_name_list:
                    entry_name = host["name_label"]
                    host_name_list.append(host["name_label"])
                else:
                    host_duplicate_count = host_name_list.count(host["name_label"])
                    entry_name = f"{host['name_label']}_{host_duplicate_count}"
                    host_name_list.append(host["name_label"])
            else:
                entry_name = host["uuid"]

            group_name = f"xo_host_{clean_group_name(host['name_label'])}"
            pool_name = self._pool_group_name_for_uuid(pools, host["$poolId"])

            self.inventory.add_group(group_name)
            self.inventory.add_host(entry_name)
            self.inventory.add_child(HOST_GROUP, entry_name)
            self.inventory.add_child(pool_name, entry_name)

            self.inventory.set_variable(entry_name, "enabled", host["enabled"])
            self.inventory.set_variable(entry_name, "hostname", host["hostname"])
            self.inventory.set_variable(entry_name, "memory", host["memory"])
            self.inventory.set_variable(entry_name, "address", host["address"])
            self.inventory.set_variable(entry_name, "cpus", host["cpus"])
            self.inventory.set_variable(entry_name, "type", "host")
            self.inventory.set_variable(entry_name, "tags", host["tags"])
            self.inventory.set_variable(entry_name, "version", host["version"])
            self.inventory.set_variable(entry_name, "power_state", host["power_state"].lower())
            self.inventory.set_variable(entry_name, "product_brand", host["productBrand"])

        for pool in pools.values():
            group_name = f"xo_pool_{clean_group_name(pool['name_label'])}"

            self.inventory.add_group(group_name)

    def _add_pools(self, pools):
        for pool in pools.values():
            group_name = f"xo_pool_{clean_group_name(pool['name_label'])}"

            self.inventory.add_group(group_name)

    # TODO: Refactor
    def _pool_group_name_for_uuid(self, pools, pool_uuid):
        for pool in pools:
            if pool == pool_uuid:
                return f"xo_pool_{clean_group_name(pools[pool_uuid]['name_label'])}"

    # TODO: Refactor
    def _host_group_name_for_uuid(self, hosts, host_uuid):
        for host in hosts:
            if host == host_uuid:
                return f"xo_host_{clean_group_name(hosts[host_uuid]['name_label'])}"

    def _populate(self, objects):
        # Prepare general groups
        self.inventory.add_group(HOST_GROUP)
        self.inventory.add_group(POOL_GROUP)
        for group in POWER_STATES:
            self.inventory.add_group(group.lower())

        self._add_pools(objects["pools"])
        self._add_hosts(objects["hosts"], objects["pools"])
        self._add_vms(objects["vms"], objects["hosts"], objects["pools"])

    def verify_file(self, path):
        valid = False
        if super().verify_file(path):
            if path.endswith(("xen_orchestra.yaml", "xen_orchestra.yml")):
                valid = True
            else:
                self.display.vvv(
                    'Skipping due to inventory source not ending in "xen_orchestra.yaml" nor "xen_orchestra.yml"'
                )
        return valid

    def parse(self, inventory, loader, path, cache=True):
        if not HAS_WEBSOCKET:
            raise AnsibleError(
                "This plugin requires websocket-client 1.0.0 or higher: "
                "https://github.com/websocket-client/websocket-client."
            )

        super().parse(inventory, loader, path)

        # read config from file, this sets 'options'
        self._read_config_data(path)
        self.inventory = inventory

        self.protocol = "wss"
        self.xoa_api_host = self.get_option("api_host")
        self.xoa_user = self.get_option("user")
        self.xoa_password = self.get_option("password")
        self.cache_key = self.get_cache_key(path)
        self.use_cache = cache and self.get_option("cache")

        self.validate_certs = self.get_option("validate_certs")
        if not self.get_option("use_ssl"):
            self.protocol = "ws"

        self.vm_entry_name_type = "uuid"
        if not self.get_option("use_vm_uuid"):
            self.vm_entry_name_type = "name_label"

        self.host_entry_name_type = "uuid"
        if not self.get_option("use_host_uuid"):
            self.host_entry_name_type = "name_label"

        objects = self._get_objects()
        self._populate(make_unsafe(objects))



# Copyright (c) Fastly, inc 2016
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: selective
type: stdout
requirements:
  - set as main display callback
short_description: Only print certain tasks
description:
  - This callback only prints tasks that have been tagged with C(print_action) or that have failed. This allows operators
    to focus on the tasks that provide value only.
  - Tasks that are not printed are placed with a C(.).
  - If you increase verbosity all tasks are printed.
options:
  nocolor:
    default: false
    description: This setting allows suppressing colorizing output.
    env:
      - name: ANSIBLE_NOCOLOR
      - name: ANSIBLE_SELECTIVE_DONT_COLORIZE
    ini:
      - section: defaults
        key: nocolor
    type: boolean
"""

EXAMPLES = r"""
- ansible.builtin.debug: msg="This will not be printed"
- ansible.builtin.debug: msg="But this will"
  tags: [print_action]
"""

import difflib

from ansible import constants as C
from ansible.module_utils.common.text.converters import to_text
from ansible.plugins.callback import CallbackBase

DONT_COLORIZE = False
COLORS = {
    "normal": "\033[0m",
    "ok": f"\x1b[{C.COLOR_CODES[C.COLOR_OK]}m",  # type: ignore
    "bold": "\033[1m",
    "not_so_bold": "\033[1m\033[34m",
    "changed": f"\x1b[{C.COLOR_CODES[C.COLOR_CHANGED]}m",  # type: ignore
    "failed": f"\x1b[{C.COLOR_CODES[C.COLOR_ERROR]}m",  # type: ignore
    "endc": "\033[0m",
    "skipped": f"\x1b[{C.COLOR_CODES[C.COLOR_SKIP]}m",  # type: ignore
}


def dict_diff(prv, nxt):
    """Return a dict of keys that differ with another config object."""
    keys = set(list(prv.keys()) + list(nxt.keys()))
    result = {}
    for k in keys:
        if prv.get(k) != nxt.get(k):
            result[k] = (prv.get(k), nxt.get(k))
    return result


def colorize(msg, color):
    """Given a string add necessary codes to format the string."""
    if DONT_COLORIZE:
        return msg
    else:
        return f"{COLORS[color]}{msg}{COLORS['endc']}"


class CallbackModule(CallbackBase):
    """selective.py callback plugin."""

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "stdout"
    CALLBACK_NAME = "community.general.selective"

    def __init__(self, display=None):
        """selective.py callback plugin."""
        super().__init__(display)
        self.last_skipped = False
        self.last_task_name = None
        self.printed_last_task = False

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        global DONT_COLORIZE
        DONT_COLORIZE = self.get_option("nocolor")

    def _print_task(self, task_name=None):
        if task_name is None:
            task_name = self.last_task_name

        if not self.printed_last_task:
            self.printed_last_task = True
            line_length = 120
            if self.last_skipped:
                self._display.display("")
            line = f"# {task_name} "
            msg = colorize(f"{line}{'*' * (line_length - len(line))}", "bold")
            self._display.display(msg)

    def _indent_text(self, text, indent_level):
        lines = text.splitlines()
        result_lines = []
        for l in lines:
            result_lines.append(f"{' ' * indent_level}{l}")
        return "\n".join(result_lines)

    def _print_diff(self, diff, indent_level):
        if isinstance(diff, dict):
            try:
                diff = "\n".join(
                    difflib.unified_diff(
                        diff["before"].splitlines(),
                        diff["after"].splitlines(),
                        fromfile=diff.get("before_header", "new_file"),
                        tofile=diff["after_header"],
                    )
                )
            except AttributeError:
                diff = dict_diff(diff["before"], diff["after"])
        if diff:
            diff = colorize(str(diff), "changed")
            self._display.display(self._indent_text(diff, indent_level + 4))

    def _print_host_or_item(self, host_or_item, changed, msg, diff, is_host, error, stdout, stderr):
        if is_host:
            indent_level = 0
            name = colorize(host_or_item.name, "not_so_bold")
        else:
            indent_level = 4
            if isinstance(host_or_item, dict):
                if "key" in host_or_item.keys():
                    host_or_item = host_or_item["key"]
            name = colorize(to_text(host_or_item), "bold")

        if error:
            color = "failed"
            change_string = colorize("FAILED!!!", color)
        else:
            color = "changed" if changed else "ok"
            change_string = colorize(f"changed={changed}", color)

        msg = colorize(msg, color)

        line_length = 120
        spaces = " " * (40 - len(name) - indent_level)
        line = f"{' ' * indent_level}  * {name}{spaces}- {change_string}"

        if len(msg) < 50:
            line += f" -- {msg}"
            self._display.display(f"{line} {'-' * (line_length - len(line))}---------")
        else:
            self._display.display(f"{line} {'-' * (line_length - len(line))}")
            self._display.display(self._indent_text(msg, indent_level + 4))

        if diff:
            self._print_diff(diff, indent_level)
        if stdout:
            stdout = colorize(stdout, "failed")
            self._display.display(self._indent_text(stdout, indent_level + 4))
        if stderr:
            stderr = colorize(stderr, "failed")
            self._display.display(self._indent_text(stderr, indent_level + 4))

    def v2_playbook_on_play_start(self, play):
        """Run on start of the play."""
        pass

    def v2_playbook_on_task_start(self, task, **kwargs):
        """Run when a task starts."""
        self.last_task_name = task.get_name()
        self.printed_last_task = False

    def _print_task_result(self, result, error=False, **kwargs):
        """Run when a task finishes correctly."""

        if "print_action" in result._task.tags or error or self._display.verbosity > 1:
            self._print_task()
            self.last_skipped = False
            msg = to_text(result._result.get("msg", "")) or to_text(result._result.get("reason", ""))

            stderr = [result._result.get("exception", None), result._result.get("module_stderr", None)]
            stderr = "\n".join([e for e in stderr if e]).strip()

            self._print_host_or_item(
                result._host,
                result._result.get("changed", False),
                msg,
                result._result.get("diff", None),
                is_host=True,
                error=error,
                stdout=result._result.get("module_stdout", None),
                stderr=stderr.strip(),
            )
            if "results" in result._result:
                for r in result._result["results"]:
                    failed = "failed" in r and r["failed"]

                    stderr = [r.get("exception", None), r.get("module_stderr", None)]
                    stderr = "\n".join([e for e in stderr if e]).strip()

                    self._print_host_or_item(
                        r[r["ansible_loop_var"]],
                        r.get("changed", False),
                        to_text(r.get("msg", "")),
                        r.get("diff", None),
                        is_host=False,
                        error=failed,
                        stdout=r.get("module_stdout", None),
                        stderr=stderr.strip(),
                    )
        else:
            self.last_skipped = True
            print(".", end="")

    def v2_playbook_on_stats(self, stats):
        """Display info about playbook statistics."""
        self._display.display("")
        self.printed_last_task = False
        self._print_task("STATS")

        hosts = sorted(stats.processed.keys())
        max_len = max((len(h) for h in hosts), default=0)
        for host in hosts:
            s = stats.summarize(host)

            if s["failures"] or s["unreachable"]:
                color = "failed"
            elif s["changed"]:
                color = "changed"
            else:
                color = "ok"

            msg = (
                f"{host.ljust(max_len)} : ok={s['ok']}\tchanged={s['changed']}\tfailed={s['failures']}\t"
                f"unreachable={s['unreachable']}\trescued={s['rescued']}\tignored={s['ignored']}"
            )
            self._display.display(colorize(msg, color))

    def v2_runner_on_skipped(self, result, **kwargs):
        """Run when a task is skipped."""
        if self._display.verbosity > 1:
            self._print_task()
            self.last_skipped = False

            line_length = 120
            spaces = " " * (31 - len(result._host.name) - 4)

            line = f"  * {colorize(result._host.name, 'not_so_bold')}{spaces}- {colorize('skipped', 'skipped')}"

            reason = result._result.get("skipped_reason", "") or result._result.get("skip_reason", "")
            if len(reason) < 50:
                line += f" -- {reason}"
                self._display.display(f"{line} {'-' * (line_length - len(line))}---------")
            else:
                self._display.display(f"{line} {'-' * (line_length - len(line))}")
                self._display.display(self._indent_text(reason, 8))
                self._display.display(reason)

    def v2_runner_on_ok(self, result, **kwargs):
        self._print_task_result(result, error=False, **kwargs)

    def v2_runner_on_failed(self, result, **kwargs):
        self._print_task_result(result, error=True, **kwargs)

    def v2_runner_on_unreachable(self, result, **kwargs):
        self._print_task_result(result, error=True, **kwargs)

    v2_playbook_on_handler_task_start = v2_playbook_on_task_start



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_serviceendpointpolicydefinition
version_added: "3.9.0"
short_description: Managed the service endpoint policy definition
description:
    - Create, update or delete the service endpoint policy definition.

options:
    resource_group:
        description:
            - The name of the resource group.
        type: str
        required: true
    service_endpoint_policy_name:
        description:
            - The name of the service endpoint policy name.
        type: str
        required: true
    name:
        description:
            - The name of the service endpoint policy definition name.
        type: str
        required: true
    description:
        description:
            - A description for this rule. Restricted to 140 chars.
        type: str
    service:
        description:
            - Service endpoint name.
        type: str
    service_resources:
        description:
            - A list of service resources.
        type: list
        elements: str
    state:
        description:
            - Set to C(present) to create or update the service endpoint policy definition.
            - Set to C(absent) to remove the service endpoint policy definition.
        default: present
        type: str
        choices:
            - absent
            - present
extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - magodo (@magodo)
    - Fred Sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Create a new service endpoint policy definition
  azure.azcollection.azure_rm_serviceendpointpolicydefinition:
    resource_group: testRG
    service_endpoint_policy_name: policy01
    name: definition01
    description: definition-test
    service: Microsoft.Storage
    service_resources:
      - "/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.Storage/storageAccounts/account01"

- name: Delete the service endpoint policy definition
  azure.azcollection.azure_rm_serviceendpointpolicydefinition:
    resource_group: testRG
    service_endpoint_policy_name: policy01
    name: definition01
    state: absent
'''

RETURN = '''
serviceendpointpolicydefinition:
    description:
        - The fact of the service endpoint policy definitions in a service end point policy.
    type: complex
    returned: always
    contains:
        id:
            description:
                - Fully qualified ID of the resource.
            type: str
            returned: always
            sample: "/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.Network/serviceEndpointPolicies/\
                     policy01/serviceEndpointPolicyDefinitions/definition01"
        name:
            description:
                - The name of the resource that is unique within a resource group.
            type: str
            returned: always
            sample: definition01
        etag:
            description:
                - A unique read-only string that changes whenever the resource is updated.
            returned: always
            type: str
            sample: "4a66fa1c-3d17-43c9-9b9d-da07ffdbb695"
        resource_group:
            description:
                - The name of the resource group.
            type: str
            returned: always
            sample: TestRG
        description:
            description:
                - A description for this rule. Restricted to 140 chars.
            type: str
            returned: always
            sample: definition-test
        service_endpoint_policy_name:
            description:
                - The name of the service endpoint policy name.
            type: str
            returned: always
            sample: policy01
        type:
            description:
                - The type of the resource.
            type: str
            returned: always
            sample: "Microsoft.Network/serviceEndpointPolicies/serviceEndpointPolicyDefinitions"
        provisioning_state:
            description:
                - The provisioning state of the service endpoint policy definition resource.
            type: str
            returned: always
            sample: Successded
        service:
            description:
                - Service endpoint name.
            type: str
            returned: always
            sample: Microsoft.Storage
        service_resources:
            description:
                - A list of service resources.
            type: list
            returned: always
            sample: ["/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.Storage/storageAccounts/account01"]
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
try:
    from azure.core.polling import LROPoller
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMServiceEndpointPolicyDefinition(AzureRMModuleBase):
    """Information class for an Azure RM Data Collection Rules"""

    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            service_endpoint_policy_name=dict(type='str', required=True),
            name=dict(type='str', required=True),
            description=dict(type='str'),
            service=dict(type='str'),
            service_resources=dict(type='list', elements='str'),
            state=dict(type='str', default='present', choices=['present', 'absent'])
        )

        self.resource_group = None
        self.service_endpoint_policy_name = None
        self.name = None
        self.description = None
        self.service = None
        self.service_resurces = None
        self.state = None
        self.log_path = None
        self.log_mode = None

        self.results = dict(
            changed=False,
            serviceendpointpolicydefinition=dict()
        )

        super(AzureRMServiceEndpointPolicyDefinition, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                                     supports_check_mode=True,
                                                                     supports_tags=False,
                                                                     facts_module=True)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        response = self.get_definition(self.name)
        changed = False
        if self.state == 'present':
            if response:
                if self.description and self.description != response.get('description'):
                    changed = True
                else:
                    self.description = response.get('description')
                if self.service and self.service != response.get('service'):
                    changed = True
                else:
                    self.service = response.get('service')
                if self.service_resources:
                    new_service_resources = response.get('service_resources', [])
                    for item in self.service_resources:
                        if item not in response.get('service_resources', []):
                            changed = True
                            new_service_resources.append(item)
                    self.service_resources = new_service_resources
                else:
                    self.service_resources = response.get('service_resources')
            else:
                changed = True
            if not self.check_mode and changed:
                response = self.create_or_update()
        else:
            if response:
                changed = True
                if self.check_mode:
                    self.log("The service endpoint policy definition already exist, will be delete")
                else:
                    response = self.delete()
            else:
                if self.check_mode:
                    self.log("There is no service endpoint policy definition.")

        self.results['changed'] = changed
        self.results['serviceendpointpolicydefinition'] = response
        return self.results

    def get_definition(self, name):
        '''
        Gets the specified service endpoint policy definition
        '''
        response = None
        try:
            response = self.network_client.service_endpoint_policy_definitions.get(self.resource_group, self.service_endpoint_policy_name, self.name)
        except Exception:
            self.log("Could not get the specified service endpoint policy definition {0}".format(self.name))
        return self.format_item(response) if response else None

    def create_or_update(self):
        '''
        Create or update the service endpoint policy definition.
        '''
        response = None
        try:
            response = self.network_client.service_endpoint_policy_definitions.begin_create_or_update(self.resource_group,
                                                                                                      self.service_endpoint_policy_name,
                                                                                                      self.name,
                                                                                                      dict(description=self.description,
                                                                                                           service=self.service,
                                                                                                           service_resources=self.service_resources))
            if isinstance(response, LROPoller):
                response = self.get_poller_result(response)
        except Exception as ex:
            self.fail("Could not create or update the service endpoint policy definition, Exception as {0}".format(ex))
        return self.format_item(response) if response else None

    def delete(self):
        '''
        Delete the service endpoint policy definition.
        '''
        try:
            response = self.network_client.service_endpoint_policy_definitions.begin_delete(self.resource_group, self.service_endpoint_policy_name, self.name)
            if isinstance(response, LROPoller):
                self.get_poller_result(response)
        except Exception as ex:
            self.fail("Could not delete the service endpoint policy definition, Exception as {0}".format(ex))

    def format_item(self, item):
        response = item.as_dict()
        response['resource_group'] = self.resource_group
        response['service_endpoint_policy_name'] = self.service_endpoint_policy_name
        return response


def main():
    """Main execution"""
    AzureRMServiceEndpointPolicyDefinition()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_compose_v2_pull

short_description: Pull a Docker compose project

version_added: 3.6.0

description:
  - Uses Docker Compose to pull images for a project.
extends_documentation_fragment:
  - community.docker._compose_v2
  - community.docker._compose_v2.minimum_version
  - community.docker._docker.cli_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
    details:
      - If O(policy=always), the module will always indicate a change. Docker Compose does not give any information whether
        pulling would update the image or not.
  diff_mode:
    support: none
  idempotent:
    support: full

options:
  policy:
    description:
      - Whether to pull images before running. This is used when C(docker compose up) is ran.
      - V(always) ensures that the images are always pulled, even when already present on the Docker daemon.
      - V(missing) only pulls them when they are not present on the Docker daemon. This is only supported since Docker Compose
        2.22.0.
    type: str
    choices:
      - always
      - missing
    default: always
  ignore_buildable:
    description:
      - If set to V(true), will not pull images that can be built.
    type: bool
    default: false
    version_added: 3.12.0
  ignore_pull_failures:
    description:
      - If set to V(true), will pull what it can and ignores images with pull failures.
    type: bool
    default: false
    version_added: 5.1.0
  include_deps:
    description:
      - If set to V(true), also pull services that are declared as dependencies.
      - This only makes sense if O(services) is used.
    type: bool
    default: false
    version_added: 3.12.0
  services:
    description:
      - Specifies a subset of services to be targeted.
    type: list
    elements: str
    version_added: 3.12.0

author:
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.docker.docker_compose_v2
"""

EXAMPLES = r"""
---
- name: Pull images for flask project
  community.docker.docker_compose_v2_pull:
    project_src: /path/to/flask
"""

RETURN = r"""
actions:
  description:
    - A list of actions that have been applied.
  returned: success
  type: list
  elements: dict
  contains:
    what:
      description:
        - What kind of resource was changed.
      type: str
      sample: container
      choices:
        - image
        - unknown
    id:
      description:
        - The ID of the resource that was changed.
      type: str
      sample: container
    status:
      description:
        - The status change that happened.
      type: str
      sample: Pulling
      choices:
        - Pulling
"""

import traceback
import typing as t

from ansible_collections.community.docker.plugins.module_utils._common_cli import (
    AnsibleModuleDockerClient,
    DockerException,
)
from ansible_collections.community.docker.plugins.module_utils._compose_v2 import (
    BaseComposeManager,
    common_compose_argspec_ex,
)
from ansible_collections.community.docker.plugins.module_utils._version import (
    LooseVersion,
)


class PullManager(BaseComposeManager):
    def __init__(self, client: AnsibleModuleDockerClient) -> None:
        super().__init__(client)
        parameters = self.client.module.params

        self.policy: t.Literal["always", "missing"] = parameters["policy"]
        self.ignore_buildable: bool = parameters["ignore_buildable"]
        self.ignore_pull_failures: bool = parameters["ignore_pull_failures"]
        self.include_deps: bool = parameters["include_deps"]
        self.services: list[str] = parameters["services"] or []

        if self.policy != "always" and self.compose_version < LooseVersion("2.22.0"):
            # https://github.com/docker/compose/pull/10981 - 2.22.0
            self.fail(
                f"A pull policy other than always is only supported since Docker Compose 2.22.0. {self.client.get_cli()} has version {self.compose_version}"
            )
        if self.ignore_buildable and self.compose_version < LooseVersion("2.15.0"):
            # https://github.com/docker/compose/pull/10134 - 2.15.0
            self.fail(
                f"--ignore-buildable is only supported since Docker Compose 2.15.0. {self.client.get_cli()} has version {self.compose_version}"
            )

    def get_pull_cmd(self, dry_run: bool) -> list[str]:
        args = self.get_base_args() + ["pull"]
        if self.policy != "always":
            args.extend(["--policy", self.policy])
        if self.ignore_buildable:
            args.append("--ignore-buildable")
        if self.ignore_pull_failures:
            args.append("--ignore-pull-failures")
        if self.include_deps:
            args.append("--include-deps")
        if dry_run:
            args.append("--dry-run")
        args.append("--")
        for service in self.services:
            args.append(service)
        return args

    def run(self) -> dict[str, t.Any]:
        result: dict[str, t.Any] = {}
        args = self.get_pull_cmd(self.check_mode)
        rc, stdout, stderr = self.client.call_cli(*args, cwd=self.project_src)
        events = self.parse_events(stderr, dry_run=self.check_mode, nonzero_rc=rc != 0)
        self.emit_warnings(events)
        self.update_result(
            result,
            events,
            stdout,
            stderr,
            ignore_service_pull_events=self.policy != "missing" and not self.check_mode,
        )
        self.update_failed(result, events, args, stdout, stderr, rc)
        self.cleanup_result(result)
        return result


def main() -> None:
    argument_spec = {
        "policy": {
            "type": "str",
            "choices": ["always", "missing"],
            "default": "always",
        },
        "ignore_buildable": {"type": "bool", "default": False},
        "ignore_pull_failures": {"type": "bool", "default": False},
        "include_deps": {"type": "bool", "default": False},
        "services": {"type": "list", "elements": "str"},
    }
    argspec_ex = common_compose_argspec_ex()
    argument_spec.update(argspec_ex.pop("argspec"))

    client = AnsibleModuleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
        needs_api_version=False,
        **argspec_ex,
    )

    try:
        manager = PullManager(client)
        result = manager.run()
        manager.cleanup()
        client.module.exit_json(**result)
    except DockerException as e:
        client.fail(
            f"An unexpected docker error occurred: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# Copyright (c) 2014-2015, Matt Martz <matt@sivel.net>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import annotations

DOCUMENTATION = r"""
author: Unknown (!UNKNOWN)
name: slack
type: notification
requirements:
  - whitelist in configuration
  - prettytable (python library)
short_description: Sends play events to a Slack channel
description:
  - This is an ansible callback plugin that sends status updates to a Slack channel during playbook execution.
options:
  http_agent:
    description:
      - HTTP user agent to use for requests to Slack.
    type: string
    version_added: "10.5.0"
  webhook_url:
    required: true
    description: Slack Webhook URL.
    type: str
    env:
      - name: SLACK_WEBHOOK_URL
    ini:
      - section: callback_slack
        key: webhook_url
  channel:
    default: "#ansible"
    description: Slack room to post in.
    type: str
    env:
      - name: SLACK_CHANNEL
    ini:
      - section: callback_slack
        key: channel
  username:
    description: Username to post as.
    type: str
    env:
      - name: SLACK_USERNAME
    default: ansible
    ini:
      - section: callback_slack
        key: username
  validate_certs:
    description: Validate the SSL certificate of the Slack server for HTTPS URLs.
    env:
      - name: SLACK_VALIDATE_CERTS
    ini:
      - section: callback_slack
        key: validate_certs
    default: true
    type: bool
"""

import json
import os
import uuid

from ansible import context
from ansible.module_utils.urls import open_url
from ansible.plugins.callback import CallbackBase

try:
    import prettytable

    HAS_PRETTYTABLE = True
except ImportError:
    HAS_PRETTYTABLE = False


class CallbackModule(CallbackBase):
    """This is an ansible callback plugin that sends status
    updates to a Slack channel during playbook execution.
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "notification"
    CALLBACK_NAME = "community.general.slack"
    CALLBACK_NEEDS_WHITELIST = True

    def __init__(self, display=None):
        super().__init__(display=display)

        if not HAS_PRETTYTABLE:
            self.disabled = True
            self._display.warning(
                "The `prettytable` python module is not installed. Disabling the Slack callback plugin."
            )

        self.playbook_name = None

        # This is a 6 character identifier provided with each message
        # This makes it easier to correlate messages when there are more
        # than 1 simultaneous playbooks running
        self.guid = uuid.uuid4().hex[:6]

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        self.webhook_url = self.get_option("webhook_url")
        self.channel = self.get_option("channel")
        self.username = self.get_option("username")
        self.show_invocation = self._display.verbosity > 1
        self.validate_certs = self.get_option("validate_certs")
        self.http_agent = self.get_option("http_agent")
        if self.webhook_url is None:
            self.disabled = True
            self._display.warning(
                "Slack Webhook URL was not provided. The "
                "Slack Webhook URL can be provided using "
                "the `SLACK_WEBHOOK_URL` environment "
                "variable."
            )

    def send_msg(self, attachments):
        headers = {
            "Content-type": "application/json",
        }

        payload = {
            "channel": self.channel,
            "username": self.username,
            "attachments": attachments,
            "parse": "none",
            "icon_url": ("https://cdn2.hubspot.net/hub/330046/file-449187601-png/ansible_badge.png"),
        }

        data = json.dumps(payload)
        self._display.debug(data)
        self._display.debug(self.webhook_url)
        try:
            response = open_url(
                self.webhook_url,
                data=data,
                validate_certs=self.validate_certs,
                headers=headers,
                http_agent=self.http_agent,
            )
            return response.read()
        except Exception as e:
            self._display.warning(f"Could not submit message to Slack: {e}")

    def v2_playbook_on_start(self, playbook):
        self.playbook_name = os.path.basename(playbook._file_name)

        title = [f"*Playbook initiated* (_{self.guid}_)"]

        invocation_items = []
        if context.CLIARGS and self.show_invocation:
            tags = context.CLIARGS["tags"]
            skip_tags = context.CLIARGS["skip_tags"]
            extra_vars = context.CLIARGS["extra_vars"]
            subset = context.CLIARGS["subset"]
            inventory = [os.path.abspath(i) for i in context.CLIARGS["inventory"]]

            invocation_items.append(f"Inventory:  {', '.join(inventory)}")
            if tags and tags != ["all"]:
                invocation_items.append(f"Tags:       {', '.join(tags)}")
            if skip_tags:
                invocation_items.append(f"Skip Tags:  {', '.join(skip_tags)}")
            if subset:
                invocation_items.append(f"Limit:      {subset}")
            if extra_vars:
                invocation_items.append(f"Extra Vars: {' '.join(extra_vars)}")

            title.append(f"by *{context.CLIARGS['remote_user']}*")

        title.append(f"\n\n*{self.playbook_name}*")
        msg_items = [" ".join(title)]
        if invocation_items:
            _inv_item = "\n".join(invocation_items)
            msg_items.append(f"```\n{_inv_item}\n```")

        msg = "\n".join(msg_items)

        attachments = [
            {
                "fallback": msg,
                "fields": [{"value": msg}],
                "color": "warning",
                "mrkdwn_in": ["text", "fallback", "fields"],
            }
        ]

        self.send_msg(attachments=attachments)

    def v2_playbook_on_play_start(self, play):
        """Display Play start messages"""

        name = play.name or f"Play name not specified ({play._uuid})"
        msg = f"*Starting play* (_{self.guid}_)\n\n*{name}*"
        attachments = [
            {
                "fallback": msg,
                "text": msg,
                "color": "warning",
                "mrkdwn_in": ["text", "fallback", "fields"],
            }
        ]
        self.send_msg(attachments=attachments)

    def v2_playbook_on_stats(self, stats):
        """Display info about playbook statistics"""

        hosts = sorted(stats.processed.keys())

        t = prettytable.PrettyTable(["Host", "Ok", "Changed", "Unreachable", "Failures", "Rescued", "Ignored"])

        failures = False
        unreachable = False

        for h in hosts:
            s = stats.summarize(h)

            if s["failures"] > 0:
                failures = True
            if s["unreachable"] > 0:
                unreachable = True

            t.add_row([h] + [s[k] for k in ["ok", "changed", "unreachable", "failures", "rescued", "ignored"]])

        attachments = []
        msg_items = [f"*Playbook Complete* (_{self.guid}_)"]
        if failures or unreachable:
            color = "danger"
            msg_items.append("\n*Failed!*")
        else:
            color = "good"
            msg_items.append("\n*Success!*")

        msg_items.append(f"```\n{t}\n```")

        msg = "\n".join(msg_items)

        attachments.append(
            {"fallback": msg, "fields": [{"value": msg}], "color": color, "mrkdwn_in": ["text", "fallback", "fields"]}
        )

        self.send_msg(attachments=attachments)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


import os
from ansible_collections.azure.azcollection.plugins.plugin_utils import file_utils
from ansible_collections.azure.azcollection.plugins.plugin_utils import connectivity_utils


class ConfigSession():
    # pylint: disable=too-many-instance-attributes
    def __init__(self, ssh_config_file, ssh_relay_file, resource_group_name, hostname, ansible_host,
                 private_key_file, local_user, port, resource_type, ssh_proxy_folder):
        self.resource_group_name = resource_group_name
        self.hostname = hostname
        self.ansible_host = ansible_host
        self.local_user = local_user
        self.port = port
        self.resource_type = resource_type
        self.proxy_path = None
        self.relay_info = None
        self.relay_info_path = None
        self.ssh_config_file = os.path.abspath(os.path.expanduser(ssh_config_file))
        self.ssh_relay_file = os.path.abspath(os.path.expanduser(ssh_relay_file))
        self.private_key_file = os.path.abspath(os.path.expanduser(private_key_file)) if private_key_file else None
        self.ssh_proxy_folder = os.path.abspath(os.path.expanduser(ssh_proxy_folder)) if ssh_proxy_folder else None

    def get_config_text(self):
        lines = [""]
        self.relay_info_path = self._create_relay_info_file()
        lines = lines + self._get_arc_entry()
        return lines

    def _get_arc_entry(self):
        lines = []
        lines.append("Host " + self.ansible_host)
        lines.append("\tHostName " + self.hostname)
        lines.append("\tUser " + self.local_user)
        if self.private_key_file:
            lines.append("\tIdentityFile \"" + self.private_key_file + "\"")
        if self.port:
            lines.append("\tProxyCommand \"" + self.proxy_path + "\" " + "-r \"" + self.relay_info_path + "\" "
                         + "-p " + str(self.port))
        else:
            lines.append("\tProxyCommand \"" + self.proxy_path + "\" " + "-r \"" + self.relay_info_path + "\"")
        return lines

    def _create_relay_info_file(self):
        relay_info_path = self.ssh_relay_file
        relay_info_dir = os.path.dirname(relay_info_path)
        if not os.path.isdir(relay_info_dir):
            os.makedirs(relay_info_dir)

        # Overwrite relay_info if it already exists in that folder.
        file_utils.delete_file(relay_info_path, f"{relay_info_path} already exists, and couldn't be overwritten.")
        file_utils.write_to_file(relay_info_path, 'w', connectivity_utils.format_relay_info_string(self.relay_info),
                                 f"Couldn't write relay information to file {relay_info_path}.", 'utf-8')
        os.chmod(relay_info_path, 0o644)

        return relay_info_path



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_inventory_devices_info
short_description: Information module for organizations _inventory _devices
description:
  - Get all organizations _inventory _devices.
  - Get organizations _inventory _devices by id.
  - Return a single device from the inventory of an organization.
  - Return the device inventory for an organization.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
  - cisco.meraki.module_info_pagination
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  perPage:
    description:
      - PerPage query parameter. The number of entries per page returned. Acceptable
        range is 3 - 1000. Default is 1000.
    type: int
  startingAfter:
    description:
      - >
        StartingAfter query parameter. A token used by the server to indicate the
        start of the page. Often this is a timestamp or an ID but it is not limited
        to those. This parameter should not be defined by client applications. The
        link for the first, last, prev, or next page in the HTTP Link header should
        define it.
    type: str
  endingBefore:
    description:
      - >
        EndingBefore query parameter. A token used by the server to indicate the end
        of the page. Often this is a timestamp or an ID but it is not limited to those.
        This parameter should not be defined by client applications. The link for
        the first, last, prev, or next page in the HTTP Link header should define
        it.
    type: str
  usedState:
    description:
      - UsedState query parameter. Filter results by used or unused inventory. Accepted
        values are 'used' or 'unused'.
    type: str
  search:
    description:
      - Search query parameter. Search for devices in inventory based on serial number,
        mac address, or model.
    type: str
  macs:
    description:
      - Macs query parameter. Search for devices in inventory based on mac addresses.
    elements: str
    type: list
  networkIds:
    description:
      - >
        NetworkIds query parameter. Search for devices in inventory based on network
        ids. Use explicit 'null' value to get available devices only.
    elements: str
    type: list
  serials:
    description:
      - Serials query parameter. Search for devices in inventory based on serials.
    elements: str
    type: list
  models:
    description:
      - Models query parameter. Search for devices in inventory based on model.
    elements: str
    type: list
  orderNumbers:
    description:
      - OrderNumbers query parameter. Search for devices in inventory based on order
        numbers.
    elements: str
    type: list
  tags:
    description:
      - >
        Tags query parameter. Filter devices by tags. The filtering is case-sensitive.
        If tags are included, 'tagsFilterType' should also be included (see below).
    elements: str
    type: list
  tagsFilterType:
    description:
      - >
        TagsFilterType query parameter. To use with 'tags' parameter, to filter devices
        which contain ANY or ALL given tags. Accepted values are 'withAnyTags' or
        'withAllTags', default is 'withAnyTags'.
    type: str
  productTypes:
    description:
      - >
        ProductTypes query parameter. Filter devices by product type. Accepted values
        are appliance, camera, campusGateway, cellularGateway, secureConnect, sensor,
        switch, systemsManager, wireless, and wirelessController.
    elements: str
    type: list
  eoxStatuses:
    description:
      - >
        EoxStatuses query parameter. Filter devices by EoX status. Accepted values
        are 'endOfSale', 'endOfSupport', 'nearEndOfSupport', or 'null'. Use 'null'
        to filter for devices with no EOX data. Supports multiple values for multi-select
        filtering.
    elements: str
    type: list
  serial:
    description:
      - Serial path parameter.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations getOrganizationInventoryDevice
    description: Complete reference of the getOrganizationInventoryDevice API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-inventory-device
  - name: Cisco Meraki documentation for organizations getOrganizationInventoryDevices
    description: Complete reference of the getOrganizationInventoryDevices API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-inventory-devices
notes:
  - SDK Method used are
    organizations.Organizations.get_organization_inventory_device,
    organizations.Organizations.get_organization_inventory_devices,
  - Paths used are
    get /organizations/{organizationId}/inventory/devices,
    get /organizations/{organizationId}/inventory/devices/{serial},
"""

EXAMPLES = r"""
- name: Get all organizations _inventory _devices
  cisco.meraki.organizations_inventory_devices_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    perPage: 0
    startingAfter: string
    endingBefore: string
    usedState: string
    search: string
    macs: []
    networkIds: []
    serials: []
    models: []
    orderNumbers: []
    tags: []
    tagsFilterType: string
    productTypes: []
    eoxStatuses: []
    organizationId: string
    total_pages: -1
    direction: next
  register: result
- name: Get organizations _inventory _devices by id
  cisco.meraki.organizations_inventory_devices_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    organizationId: string
    serial: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "mac": "string",
      "serial": "string",
      "name": "string",
      "model": "string",
      "networkId": "string",
      "orderNumber": "string",
      "claimedAt": "string",
      "licenseExpirationDate": "string",
      "tags": [
        "string"
      ],
      "productType": "string",
      "countryCode": "string",
      "details": [
        {
          "name": "string",
          "value": "string"
        }
      ],
      "eox": {
        "status": "string",
        "endOfSaleAt": "string",
        "endOfSupportAt": "string"
      }
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2024 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_networkwatcher_info

version_added: "2.5.0"

short_description: Get or list the network watcher facts

description:
    - Get or list the network watcher facts.

options:
    resource_group:
        description:
            - The name of the resource group.
        type: str
    name:
        description:
            - Name of the network watcher.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)

'''

EXAMPLES = '''
- name: Get the network watcher facts
  azure_rm_networkwatcher_info:
    resource_group: myResourceGroup
    name: mywatcher01

- name: list the network watcher facts
  azure_rm_networkwatcher_info:
    resource_group: myResourceGroup

- name: list the network watcher and filter by tags
  azure_rm_networkwatcher_info:
    tags:
      - key1
      - key2
'''

RETURN = '''
network_watchers:
    description:
        - The facts of the network watcher.
    returned: always
    type: complex
    contains:
        resource_group:
            description:
                - The resource group.
            type: str
            returned: always
            sample: NetworkWatcherRG
        id:
            description:
                - Resource ID.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx/resourceGroups/NetworkWatcherRG/providers/Microsoft.Network/networkWatchers/netwatcher_eastus"
        location:
            description:
                - Resource location.
            returned: always
            type: str
            sample: eastus
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: mynetworkwatcher01
        tags:
            description:
                - Resource tags.
            returned: always
            type: dict
            sample: { 'key1':'value1' }
        type:
            description:
                - Resource type.
            returned: always
            type: str
            sample: "Microsoft.Network/networkWatchers"
        provisioning_state:
            description:
                - The provisioning state of the network watcher resource.
            type: str
            returned: always
            sample: Succeeded
'''

try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.mgmt.core.tools import parse_resource_id
except Exception:
    # This is handled in azure_rm_common
    pass

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMNetworkWatcherInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            resource_group=dict(type='str'),
            name=dict(type='str'),
            tags=dict(type='list', elements='str')
        )

        self.results = dict(
            changed=False,
            network_watchers=[]
        )

        self.resource_group = None
        self.name = None
        self.tags = None

        super(AzureRMNetworkWatcherInfo, self).__init__(self.module_arg_spec,
                                                        supports_check_mode=True,
                                                        supports_tags=False,
                                                        facts_module=True)

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.name and self.resource_group:
            response = [self.get_by_name()]
        elif self.resource_group:
            response = self.list_by_resourcegroup()
        else:
            response = self.list_all()

        self.results['network_watchers'] = [self.to_dict(item) for item in response if response is not None]

        return self.results

    def get_by_name(self):
        response = None
        try:
            response = self.network_client.network_watchers.get(self.resource_group, self.name)

        except ResourceNotFoundError as exec:
            self.log("Failed to get network watchers, Exception as {0}".format(exec))

        return response

    def list_by_resourcegroup(self):
        response = None
        try:
            response = self.network_client.network_watchers.list(self.resource_group)
        except Exception as exec:
            self.log("Faild to list network watchers by resource group, exception as {0}".format(exec))
        return response

    def list_all(self):
        response = None
        try:
            response = self.network_client.network_watchers.list_all()
        except Exception as exc:
            self.fail("Failed to list all items - {0}".format(str(exc)))

        return response

    def to_dict(self, body):
        results = dict()
        if body is not None and self.has_tags(body.tags, self.tags):
            results = dict(
                resource_group=parse_resource_id(body.id).get('resource_group'),
                id=body.id,
                name=body.name,
                location=body.location,
                tags=body.tags,
                type=body.type,
                provisioning_state=body.provisioning_state,
            )
            return results
        return None


def main():
    AzureRMNetworkWatcherInfo()


if __name__ == '__main__':
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The ios_vrf_global config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.facts.facts import Facts
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.rm_templates.vrf_global import (
    Vrf_globalTemplate,
)


class Vrf_global(ResourceModule):
    """
    The ios_vrf_global config class
    """

    def __init__(self, module):
        super(Vrf_global, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="vrf_global",
            tmplt=Vrf_globalTemplate(),
        )
        self.parsers = [
            "description",
            "ipv4.multicast.multitopology",
            "ipv6.multicast.multitopology",
            "rd",
            "vnet.tag",
            "vpn.id",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        haved, wantd = dict(), dict()
        self.want = self._handle_deprecates(self.want)

        if self.want:
            for entry in self.want.get("vrfs", []):
                wantd.update({(entry["name"]): entry})

        if self.have:
            for entry in self.have.get("vrfs", []):
                haved.update({(entry["name"]): entry})

        # if state is merged, merge want onto have
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, limit the have to anything in want & set want to nothing
        if self.state == "deleted":
            haved = {k: v for k, v in haved.items() if k in wantd or not wantd}
            wantd = {}

        if self.state in ["overridden", "deleted"]:
            for k, have in haved.items():
                if k not in wantd:
                    self._compare(want={}, have=have)

        if self.state == "purged":
            for k, have in haved.items():
                self.purge(have)

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Vrf_global network resource.
        """
        wanted = deepcopy(want)
        haved = deepcopy(have)

        rt_want = wanted.get("route_target", {})
        if rt_want:
            if rt_want.get("exports"):
                rt_want["exports"].sort()
            if rt_want.get("imports"):
                rt_want["imports"].sort()

        rt_have = haved.get("route_target", {})
        if rt_have:
            if rt_have.get("exports"):
                rt_have["exports"].sort()
            if rt_have.get("imports"):
                rt_have["imports"].sort()

        if wanted == haved:
            return

        self.addcmd(want or have, "name", False)
        self.compare(self.parsers, want, have)
        self._compare_route_targets(want, have)

    def _compare_route_targets(self, want, have):
        """Specialized comparison for route-target lists using set logic."""
        rt_want = want.get("route_target", {}) or {}
        rt_have = have.get("route_target", {}) or {}

        # Compare Exports
        want_exports = set(rt_want.get("exports") or [])
        have_exports = set(rt_have.get("exports") or [])

        for item in want_exports - have_exports:
            self.addcmd({"item": item}, "route_target.exports", negate=False)
        for item in have_exports - want_exports:
            self.addcmd({"item": item}, "route_target.exports", negate=True)

        # Compare Imports
        want_imports = set(rt_want.get("imports") or [])
        have_imports = set(rt_have.get("imports") or [])

        for item in want_imports - have_imports:
            self.addcmd({"item": item}, "route_target.imports", negate=False)
        for item in have_imports - want_imports:
            self.addcmd({"item": item}, "route_target.imports", negate=True)

    def purge(self, have):
        """Purge the VRF configuration"""
        self.commands.append("no vrf definition {0}".format(have["name"]))

    def _handle_deprecates(self, want):
        if not isinstance(want, dict) or "vrfs" not in want:
            return want
        for vrf_config in want["vrfs"]:
            if "route_target" in vrf_config:
                rt = vrf_config["route_target"]
                if "exports" in rt and isinstance(rt["exports"], str):
                    rt["exports"] = [rt["exports"]]
                if "imports" in rt and isinstance(rt["imports"], str):
                    rt["imports"] = [rt["imports"]]
        return want



#!/usr/bin/python

# Copyright (c) 2017, Kairo Araujo <kairo@kairo.eti.br>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: installp
author:
  - Kairo Araujo (@kairoaraujo)
short_description: Manage packages on AIX
description:
  - Manage packages using 'installp' on AIX.
deprecated:
  removed_in: 15.0.0
  why: The module is not actively maintained.
  alternative: >-
    Use C(ibm.power_aix.installp) instead.
    See U(https://ibm.github.io/ansible-power-aix/modules/installp.html) for details.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  accept_license:
    description:
      - Whether to accept the license for the package(s).
    type: bool
    default: false
  name:
    description:
      - One or more packages to install or remove.
      - Use V(all) to install all packages available on informed O(repository_path).
    type: list
    elements: str
    required: true
    aliases: [pkg]
  repository_path:
    description:
      - Path with AIX packages (required to install).
    type: path
  state:
    description:
      - Whether the package needs to be present on or absent from the system.
    type: str
    choices: [absent, present]
    default: present
notes:
  - If the package is already installed, even the package/fileset is new, the module does not install it.
"""

EXAMPLES = r"""
- name: Install package foo
  community.general.installp:
    name: foo
    repository_path: /repository/AIX71/installp/base
    accept_license: true
    state: present

- name: Install bos.sysmgt that includes bos.sysmgt.nim.master, bos.sysmgt.nim.spot
  community.general.installp:
    name: bos.sysmgt
    repository_path: /repository/AIX71/installp/base
    accept_license: true
    state: present

- name: Install bos.sysmgt.nim.master only
  community.general.installp:
    name: bos.sysmgt.nim.master
    repository_path: /repository/AIX71/installp/base
    accept_license: true
    state: present

- name: Install bos.sysmgt.nim.master and bos.sysmgt.nim.spot
  community.general.installp:
    name: bos.sysmgt.nim.master, bos.sysmgt.nim.spot
    repository_path: /repository/AIX71/installp/base
    accept_license: true
    state: present

- name: Remove packages bos.sysmgt.nim.master
  community.general.installp:
    name: bos.sysmgt.nim.master
    state: absent
"""

RETURN = r""" # """

import os
import re

from ansible.module_utils.basic import AnsibleModule


def _check_new_pkg(module, package, repository_path):
    """
    Check if the package of fileset is correct name and repository path.

    :param module: Ansible module arguments spec.
    :param package: Package/fileset name.
    :param repository_path: Repository package path.
    :return: Bool, package information.
    """

    if os.path.isdir(repository_path):
        installp_cmd = module.get_bin_path("installp", True)
        rc, package_result, err = module.run_command([installp_cmd, "-l", "-MR", "-d", repository_path])
        if rc != 0:
            module.fail_json(msg="Failed to run installp.", rc=rc, err=err)

        if package == "all":
            pkg_info = "All packages on dir"
            return True, pkg_info

        else:
            pkg_info = {}
            for line in package_result.splitlines():
                if re.findall(package, line):
                    pkg_name = line.split()[0].strip()
                    pkg_version = line.split()[1].strip()
                    pkg_info[pkg_name] = pkg_version

                    return True, pkg_info

        return False, None

    else:
        module.fail_json(msg=f"Repository path {repository_path} is not valid.")


def _check_installed_pkg(module, package, repository_path):
    """
    Check the package on AIX.
    It verifies if the package is installed and information

    :param module: Ansible module parameters spec.
    :param package: Package/fileset name.
    :param repository_path: Repository package path.
    :return: Bool, package data.
    """

    lslpp_cmd = module.get_bin_path("lslpp", True)
    rc, lslpp_result, err = module.run_command([lslpp_cmd, "-lcq", f"{package}*"])

    if rc == 1:
        package_state = " ".join(err.split()[-2:])
        if package_state == "not installed.":
            return False, None
        else:
            module.fail_json(msg="Failed to run lslpp.", rc=rc, err=err)

    if rc != 0:
        module.fail_json(msg="Failed to run lslpp.", rc=rc, err=err)

    pkg_data = {}
    full_pkg_data = lslpp_result.splitlines()
    for line in full_pkg_data:
        pkg_name, fileset, level = line.split(":")[0:3]
        pkg_data[pkg_name] = fileset, level

    return True, pkg_data


def remove(module, installp_cmd, packages):
    repository_path = None
    remove_count = 0
    removed_pkgs = []
    not_found_pkg = []
    for package in packages:
        pkg_check, dummy = _check_installed_pkg(module, package, repository_path)

        if pkg_check:
            if not module.check_mode:
                rc, remove_out, err = module.run_command([installp_cmd, "-u", package])
                if rc != 0:
                    module.fail_json(msg="Failed to run installp.", rc=rc, err=err)
            remove_count += 1
            removed_pkgs.append(package)

        else:
            not_found_pkg.append(package)

    if remove_count > 0:
        if len(not_found_pkg) > 1:
            not_found_pkg.insert(0, "Package(s) not found: ")

        changed = True
        msg = f"Packages removed: {' '.join(removed_pkgs)}. {' '.join(not_found_pkg)} "

    else:
        changed = False
        msg = f"No packages removed, all packages not found: {' '.join(not_found_pkg)}"

    return changed, msg


def install(module, installp_cmd, packages, repository_path, accept_license):
    installed_pkgs = []
    not_found_pkgs = []
    already_installed_pkgs = {}

    accept_license_param = {
        True: ["-Y"],
        False: [],
    }

    # Validate if package exists on repository path.
    for package in packages:
        pkg_check, pkg_data = _check_new_pkg(module, package, repository_path)

        # If package exists on repository path, check if package is installed.
        if pkg_check:
            pkg_check_current, pkg_info = _check_installed_pkg(module, package, repository_path)

            # If package is already installed.
            if pkg_check_current:
                # Check if package is a package and not a fileset, get version
                # and add the package into already installed list
                if package in pkg_info.keys():
                    already_installed_pkgs[package] = pkg_info[package][1]

                else:
                    # If the package is not a package but a fileset, confirm
                    # and add the fileset/package into already installed list
                    for key in pkg_info.keys():
                        if package in pkg_info[key]:
                            already_installed_pkgs[package] = pkg_info[key][1]

            else:
                if not module.check_mode:
                    rc, out, err = module.run_command(
                        [installp_cmd, "-a"]
                        + accept_license_param[accept_license]
                        + ["-X", "-d", repository_path, package]
                    )
                    if rc != 0:
                        module.fail_json(msg="Failed to run installp", rc=rc, err=err)
                installed_pkgs.append(package)

        else:
            not_found_pkgs.append(package)

    if len(installed_pkgs) > 0:
        installed_msg = f" Installed: {' '.join(installed_pkgs)}."
    else:
        installed_msg = ""

    if len(not_found_pkgs) > 0:
        not_found_msg = f" Not found: {' '.join(not_found_pkgs)}."
    else:
        not_found_msg = ""

    if len(already_installed_pkgs) > 0:
        already_installed_msg = f" Already installed: {already_installed_pkgs}."
    else:
        already_installed_msg = ""

    if len(installed_pkgs) > 0:
        changed = True
        msg = f"{installed_msg}{not_found_msg}{already_installed_msg}"
    else:
        changed = False
        msg = f"No packages installed.{installed_msg}{not_found_msg}{already_installed_msg}"

    return changed, msg


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="list", elements="str", required=True, aliases=["pkg"]),
            repository_path=dict(type="path"),
            accept_license=dict(type="bool", default=False),
            state=dict(type="str", default="present", choices=["absent", "present"]),
        ),
        supports_check_mode=True,
    )

    name = module.params["name"]
    repository_path = module.params["repository_path"]
    accept_license = module.params["accept_license"]
    state = module.params["state"]

    installp_cmd = module.get_bin_path("installp", True)

    if state == "present":
        if repository_path is None:
            module.fail_json(msg="repository_path is required to install package")

        changed, msg = install(module, installp_cmd, name, repository_path, accept_license)

    elif state == "absent":
        changed, msg = remove(module, installp_cmd, name)

    else:
        module.fail_json(changed=False, msg="Unexpected state.")

    module.exit_json(changed=changed, msg=msg)


if __name__ == "__main__":
    main()
